"""
Addition.
"""
from scipy.misc import comb
import numpy

from chaospy.dist import Dist
import chaospy.dist.operators


class Add(Dist):
    """Addition."""

    def __init__(self, left, right):
        """
        Constructor.

        Args:
            left (Dist, array_like) : Left hand side.
            right (Dist, array_like) : Right hand side.
        """
        left_ = not isinstance(left, Dist) or 1 and len(left)
        right_ = not isinstance(right, Dist) or 1 and len(right)
        length = max(left_, right_)

        Dist.__init__(
            self, left=left, right=right, _length=length,
            _advance=True)

    def _str(self, left, right):
        """String representation."""
        return "(%s+%s)" % (left, right)

    def _val(self, graph):
        """Value extraction."""
        if len(graph.keys) == 2:
            return graph.keys["left"] + graph.keys["right"]
        return self

    def _bnd(self, xloc, graph):
        """Distribution bounds."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
        lower, upper = graph((xloc.T-num.T).T, dist)
        return (lower.T+num.T).T, (upper.T+num.T).T

    def _cdf(self, xloc, graph):
        """Cumulative distribution function."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
        return graph((xloc.T-num.T).T, dist)

    def _pdf(self, xloc, graph):
        """Probability density function."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
        return graph((xloc.T-num.T).T, dist)

    def _ppf(self, uloc, graph):
        """Point percentile function."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
        return (graph(uloc, dist).T + num.T).T

    def _mom(self, keys, graph):
        """Statistical moments."""
        if len(graph.dists) == 2 and \
                graph.dists["left"].dependent(graph.dists["right"]):
            raise NotImplementedError("dependency")

        dim = len(self)
        kmax = numpy.max(keys, 1)+1
        keys_ = numpy.mgrid[[slice(0, _, 1) for _ in kmax]]
        keys_ = keys_.reshape(dim, int(keys_.size/dim))

        left = []
        if "left" in graph.dists:
            left.append(graph(keys_, graph.dists["left"]))
        else:
            left.append(graph.keys["left"]**keys_.T)
        if "right" in graph.dists:
            left.append(graph(keys_, graph.dists["right"]))
        else:
            left.append(graph.keys["right"]**keys_.T)

        if dim == 1:
            left[0] = left[0].flatten()
            left[1] = left[1].flatten()
            keys = keys.flatten()

        out = 0.
        for idx in range(keys_.shape[1]):
            key = keys_.T[idx]
            coef = comb(keys.T, key)
            out = out + coef*left[0][idx]*left[1][keys-idx]*(key <= keys.T)

        if dim > 1:
            out = numpy.prod(out, 1)
        return out

    def _ttr(self, order, graph):
        """Three terms recursion coefficients."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]

        coeffs0, coeffs1 = graph(order, dist)
        return coeffs0+num, coeffs1


def add(left, right):
    """
    Distribution addition.

    Args:
        left (Dist, array_like) : left hand side.
        right (Dist, array_like) : right hand side.
    """
    if left is right:
        return chaospy.dist.operators.mul(2, left)

    if isinstance(left, Dist):
        if not isinstance(right, Dist):
            right = numpy.array(right)
            if right.size == 1 and right == 0:
                return left

    elif isinstance(right, Dist):
        left = numpy.array(left)
        if left.size == 1 and left == 0:
            return right

    else:
        return left+right

    return Add(left=left, right=right)



"""
Negative of a distribution.
"""
import numpy
from chaospy.dist import Dist

class Neg(Dist):
    """Negative of a distribution."""

    def __init__(self, left):
        """
        Constructor.

        Args:
            left (Dist) : distribution.
        """
        Dist.__init__(self, left=left,
                _length=len(left), _advance=True)

    def _bnd(self, xloc, graph):
        """Distribution bounds."""
        return -graph(-xloc, graph.dists["left"])[::-1]

    def _pdf(self, xloc, graph):
        """Probability density function."""
        return graph(-xloc, graph.dists["left"])

    def _cdf(self, xloc, graph):
        """Cumulative distribution function."""
        return 1-graph(-xloc, graph.dists["left"])

    def _ppf(self, q, graph):
        """Point percentile function."""
        return -graph(1-q, graph.dists["left"])

    def _mom(self, k, graph):
        """Statistical moments."""
        return (-1)**numpy.sum(k, 0)*graph(k, graph.dists["left"])

    def _ttr(self, k, graph):
        """Three terms recursion coefficients."""
        a,b = graph(k, graph.dists["left"])
        return -a,b

    def _str(self, left):
        """String representation."""
        return "(-" + str(left) + ")"

    def _val(self, graph):
        """Value extraction."""
        if "left" in graph.keys:
            return -graph.keys["left"]
        return self

def neg(left):
    """
    Negative of a distribution.

    Args:
        dist (Dist) : distribution.
    """
    if not isinstance(left, Dist):
        return -left
    return Neg(left)




"""
Point collcation method, or regression based polynomial chaos expansion builds
open the idea of fitting a polynomial chaos expansion to a set of generated
samples and evaluations. The experiement can be done as follows:

- Select a :ref:`distributions`::

      >>> distribution = cp.Iid(cp.Normal(0, 1), 2)

- Generate :ref:`orthogonality`::

      >>> orthogonal_expansion = cp.orth_ttr(2, distribution)
      >>> print(orthogonal_expansion)
      [1.0, q1, q0, q1^2-1.0, q0q1, q0^2-1.0]

- Generate samples using :ref:`montecarlo` (or alternative absissas from
  :ref:`quadrature`)::

      >>> samples = distribution.sample(
      ...     2*len(orthogonal_expansion), rule="M")
      >>> print(samples[:, :4])
      [[-1.42607687 -1.02007623 -0.73631592 -0.50240222]
       [ 0.         -0.67448975  0.67448975 -1.15034938]]

- A function evaluated using the nodes generated in the second step::

      >>> def model_solver(param):
      ...     return [param[0]*param[1], param[0]*np.e**-param[1]+1]
      >>> solves = [model_solver(sample) for sample in samples.T]
      >>> print(np.around(solves[:4], 8))
      [[-0.         -0.42607687]
       [ 0.68803096 -1.00244135]
       [-0.49663754  0.62490868]
       [ 0.57793809 -0.58723759]]

- Bring it all together using `~chaospy.collocation.fit_regression`::

      >>> approx_model = cp.fit_regression(
      ...      orthogonal_expansion, samples, solves)
      >>> print(cp.around(approx_model, 5))
      [q0q1, 0.15275q0^2-1.23005q0q1+0.16104q1^2+1.214q0+0.044q1+0.86842]

In this example, the number of collocation points is selected to be twice the
number of unknown coefficients :math:`N+1`. This
follows the default outlined in :cite:`hosder_efficient_2007`. Changing this is
obviously possible. When the number of parameter is equal the number of
unknown, the, the polynomial approximation becomes an interpolation method and
overlap with Lagrange polynomials. If the number of samples are fewer than the
number of unknown, classical least squares can not be used.  Instead it
possible to use methods for doing estimation with too few samples.

The function :func:`~chaospy.collocation.fit_regression` also takes an optional
``rule`` keyword argument. It allows for the selection of regression method
used when fitting the samples to the polynomials.  One example there is
orthogonal matching pursuit :cite:`mallat_matching_1993`. It forces the
result to have at most one non-zero coefficient. To implement it use the
keyword ``rule="OMP"``, and to force the number of coefficients to be
for example 1: ``n_nonzero_coefs=1``. In practice::

   >>> approx_model = cp.fit_regression(
   ...     orthogonal_expansion, samples, solves,
   ...     rule="OMP", n_nonzero_coefs=1)
   >>> print(cp.around(approx_model, 8))
   [q0q1, 1.52536468q0]

Except for least squares, Tikhonov regularization with and without cross
validation, all the method listed is taken from ``sklearn`` software. All
optional arguments for various methods is covered in both
``sklearn.linear_model`` and in ``cp.fit_regression``.

The follwong methods uses scikits-learn as backend.
See `sklearn.linear_model` for more details.

+--------+---------------------------+-------------------------------------+
| Key    | Scikit-learn name         | Description                         |
|        | Parameters                |                                     |
+========+===========================+=====================================+
| "BARD" | ARDRegression             | Bayesian ARD Regression             |
|        | n_iter=300                | Maximum iterations                  |
|        | tol=1e-3                  | Optimization tolerance              |
|        | alpha_1=1e-6              | Gamma scale parameter               |
|        | alpha_2=1e-6              | Gamma inverse scale parameter       |
|        | lambda_1=1e-6             | Gamma shape parameter               |
|        | lambda_2=1e-6             | Gamma inverse scale parameter       |
|        | threshold_lambda=1e-4     | Upper pruning threshold             |
+--------+---------------------------+-------------------------------------+
| "BR"   | BayesianRidge             | Bayesian Ridge Regression           |
|        | n_iter=300                | Maximum iterations                  |
|        | tol=1e-3                  | Optimization tolerance              |
|        | alpha_1=1e-6              | Gamma scale parameter               |
|        | alpha_2=1e-6              | Gamma inverse scale parameter       |
|        | lambda_1=1e-6             | Gamma shape parameter               |
|        | lambda_2=1e-6             | Gamma inverse scale parameter       |
+--------+---------------------------+-------------------------------------+
| "EN"   | ElastiNet                 | Elastic Net                         |
|        | alpha=1.0                 | Dampening parameter                 |
|        | rho                       | Mixing parameter in [0,1]           |
|        | max_iter=300              | Maximum iterations                  |
|        | tol                       | Optimization tolerance              |
+--------+---------------------------+-------------------------------------+
| "ENC"  | ElasticNetCV              | EN w/Cross Validation               |
|        | rho                       | Dampening parameter(s)              |
|        | eps=1e-3                  | min(alpha)/max(alpha)               |
|        | n_alphas                  | Number of alphas                    |
|        | alphas                    | List of alphas                      |
|        | max_iter                  | Maximum iterations                  |
|        | tol                       | Optimization tolerance              |
|        | cv=3                      | Cross validation folds              |
+--------+---------------------------+-------------------------------------+
| "LA"   | Lars                      | Least Angle Regression              |
|        | n_nonzero_coefs           | Number of non-zero coefficients     |
|        | eps                       | Cholesky regularization             |
+--------+---------------------------+-------------------------------------+
| "LAC"  | LarsCV                    | LAR w/Cross Validation              |
|        | max_iter                  | Maximum iterations                  |
|        | cv=5                      | Cross validation folds              |
|        | max_n_alphas              | Max points for residuals in cv      |
+--------+---------------------------+-------------------------------------+
| "LAS"  | Lasso                     | Least Abs Shrink \& Select Operator |
|        | alpha=1.0                 | Dampening parameter                 |
|        | max_iter                  | Maximum iterations                  |
|        | tol                       | Optimization tolerance              |
+--------+---------------------------+-------------------------------------+
| "LASC" | LassoCV                   | LAS w/Cross Validation              |
|        | eps=1e-3                  | min(alpha)/max(alpha)               |
|        | n_alphas                  | Number of alphas                    |
|        | alphas                    | List of alphas                      |
|        | max_iter                  | Maximum iterations                  |
|        | tol                       | Optimization tolerance              |
|        | cv=3                      | Cross validation folds              |
+--------+---------------------------+-------------------------------------+
| "LL"   | LassoLars                 | Lasso and Lars model                |
|        | max_iter                  | Maximum iterations                  |
|        | eps                       | Cholesky regularization             |
+--------+---------------------------+-------------------------------------+
| "LLC"  | LassoLarsCV               | LL w/Cross Validation               |
|        | max_iter                  | Maximum iterations                  |
|        | cv=5                      | Cross validation folds              |
|        | max_n_alphas              | Max points for residuals in cv      |
|        | eps                       | Cholesky regularization             |
+--------+---------------------------+-------------------------------------+
| "LLIC" | LassoLarsIC               | LL w/AIC or BIC                     |
|        | criterion                 | "AIC" or "BIC" criterion            |
|        | max_iter                  | Maximum iterations                  |
|        | eps                       | Cholesky regularization             |
+--------+---------------------------+-------------------------------------+
| "OMP"  | OrthogonalMatchingPursuit | Orthogonal matching pursuit         |
|        | n_nonzero_coefs           | Number of non-zero coefficients     |
|        | tol                       | Max residual norm                   |
+--------+---------------------------+-------------------------------------+

There is also the following local methods:

+------------+----------------------------------------------+
| Key        | Description                                  |
| Parameters |                                              |
+============+==============================================+
| "LS"       | Ordenary Least Squares                       |
+------------+----------------------------------------------+
| "T"        | Ridge Regression/Tikhonov Regularization     |
| order      | Order of regularization (or custom matrix)   |
| alpha      | Dampning parameter (else estimated from gcv) |
+------------+----------------------------------------------+
| "TC"       | Tikhonov with Cross Validation               |
| order      | Order of regularization (or custom matrix)   |
| alpha      | Dampning parameter (else estimated from gcv) |
+------------+----------------------------------------------+
"""

__all__ = [
"fit_regression", "lstsq_cv", "rlstsq"
]

import numpy as np
from scipy import linalg, optimize

try:
    from sklearn import linear_model
except:
    pass

import chaospy as cp

def fit_regression(P, x, u, rule="LS", retall=False, **kws):
    """
    Fit a polynomial chaos expansion using linear regression.

    Args:
        P (Poly) : Polynomial expansion with `P.shape=(M,)` and `P.dim=D`.
        x (array_like) : Collocation nodes with `x.shape=(D,K)`.
        u (array_like) : Model evaluations with `len(u)=K`.
        retall (bool) : If True return Fourier coefficients in addition to R.
        rule (str) : Regression method used.

    Returns:
        (Poly, np.ndarray) : Fitted polynomial with `R.shape=u.shape[1:]` and
                `R.dim=D`. The Fourier coefficients in the estimation.

    Examples:
        >>> x, y = cp.variable(2)
        >>> P = cp.Poly([1, x, y])
        >>> s = [[-1,-1,1,1], [-1,1,-1,1]]
        >>> u = [0,1,1,2]
        >>> print(cp.around(cp.fit_regression(P, s, u), 14))
        0.5q0+0.5q1+1.0
    """
    x = np.array(x)
    if len(x.shape)==1:
        x = x.reshape(1, *x.shape)
    u = np.array(u)

    Q = P(*x).T
    shape = u.shape[1:]
    u = u.reshape(u.shape[0], int(np.prod(u.shape[1:])))

    rule = rule.upper()

    # Local rules
    if rule=="LS":
        uhat = linalg.lstsq(Q, u)[0].T

    elif rule=="T":
        uhat, alphas = rlstsq(Q, u, kws.get("order",0),
                kws.get("alpha", None), False, True)
        uhat = uhat.T

    elif rule=="TC":
        uhat = rlstsq(Q, u, kws.get("order",0),
                kws.get("alpha", None), True)
        uhat = uhat.T

    else:

        # Scikit-learn wrapper
        try:
            _ = linear_model
        except:
            raise NotImplementedError(
                    "sklearn not installed")

        if rule=="BARD":
            solver = linear_model.ARDRegression(
                fit_intercept=False, copy_X=False, **kws)

        elif rule=="BR":
            kws["fit_intercept"] = kws.get("fit_intercept", False)
            solver = linear_model.BayesianRidge(**kws)

        elif rule=="EN":
            kws["fit_intercept"] = kws.get("fit_intercept", False)
            solver = linear_model.ElasticNet(**kws)

        elif rule=="ENC":
            kws["fit_intercept"] = kws.get("fit_intercept", False)
            solver = linear_model.ElasticNetCV(**kws)

        elif rule=="LA": # success
            kws["fit_intercept"] = kws.get("fit_intercept", False)
            solver = linear_model.Lars(**kws)

        elif rule=="LAC":
            kws["fit_intercept"] = kws.get("fit_intercept", False)
            solver = linear_model.LarsCV(**kws)

        elif rule=="LAS":
            kws["fit_intercept"] = kws.get("fit_intercept", False)
            solver = linear_model.Lasso(**kws)

        elif rule=="LASC":
            kws["fit_intercept"] = kws.get("fit_intercept", False)
            solver = linear_model.LassoCV(**kws)

        elif rule=="LL":
            kws["fit_intercept"] = kws.get("fit_intercept", False)
            solver = linear_model.LassoLars(**kws)

        elif rule=="LLC":
            kws["fit_intercept"] = kws.get("fit_intercept", False)
            solver = linear_model.LassoLarsCV(**kws)

        elif rule=="LLIC":
            kws["fit_intercept"] = kws.get("fit_intercept", False)
            solver = linear_model.LassoLarsIC(**kws)

        elif rule=="OMP":
            solver = linear_model.OrthogonalMatchingPursuit(**kws)

        uhat = solver.fit(Q, u).coef_

    u = u.reshape(u.shape[0], *shape)

    R = cp.poly.sum((P*uhat), -1)
    R = cp.poly.reshape(R, shape)

    if retall==1:
        return R, uhat

    elif retall==2:
        if rule=="T":
            return R, uhat, Q, alphas
        return R, uhat, Q

    return R


def rlstsq(A, b, order=1, alpha=None, cross=False, retall=False):
    """
    Least Squares Minimization using Tikhonov regularization.

    Includes method for robust generalized cross-validation.

    Args:
        A (array_like, shape (M,N)) : "Coefficient" matrix.
        b (array_like, shape (M,) or (M, K)) : Ordinate or "dependent
                variable" values. If `b` is two-dimensional, the least-squares
                solution is calculated for each of the `K` columns of `b`.
        order (int, array_like) : If int, it is the order of Tikhonov
                regularization.  If array_like, it will be used as
                regularization matrix.
        alpha (float, optional) : Lower threshold for the dampening parameter.
                The real value is calculated using generalised cross
                validation.
        cross (bool) : Use cross validation
        retall (bool) : If True, return also estimated alpha-value
    """
    A = np.array(A)
    b = np.array(b)
    m,l = A.shape

    if cross:
        out = np.empty((m,l) + b.shape[1:])
        A_ = np.empty((m-1,l))
        b_ = np.empty((m-1,) + b.shape[1:])
        for i in range(m):
            A_[:i] = A[:i]
            A_[i:] = A[i+1:]
            b_[:i] = b[:i]
            b_[i:] = b[i+1:]
            out[i] = rlstsq(A_, b_, order, alpha, False)

        return np.median(out, 0)

    if order==0:
        L = np.eye(l)

    elif order==1:
        L = np.zeros((l-1,l))
        L[:,:-1] -= np.eye(l-1)
        L[:,1:] += np.eye(l-1)

    elif order==2:
        L = np.zeros((l-2,l))
        L[:,:-2] += np.eye(l-2)
        L[:,1:-1] -= 2*np.eye(l-2)
        L[:,2:] += np.eye(l-2)

    elif order is None:
        L = np.zeros(1)

    else:
        L = np.array(order)
        assert L.shape[-1]==l or L.shape in ((), (1,))

    if alpha is None and not (order is None):

        gamma = 0.1

        def rgcv_error(alpha):
            if alpha<=0: return np.inf
            A_ = np.dot(A.T,A)+alpha*(np.dot(L.T,L))
            try:
                A_ = np.dot(linalg.inv(A_), A.T)
            except linalg.LinAlgError:
                return np.inf
            x = np.dot(A_, b)
            res2 = np.sum((np.dot(A,x)-b)**2)
            K = np.dot(A, A_)
            V = m*res2/np.trace(np.eye(m)-K)**2
            mu2 = np.sum(K*K.T)/m

            return (gamma + (1-gamma)*mu2)*V

        alpha = optimize.fmin(rgcv_error, 1, disp=0)

    out = linalg.inv(np.dot(A.T,A) + alpha*np.dot(L.T, L))
    out = np.dot(out, np.dot(A.T, b))
    if retall:
        return out, alpha
    return out


def lstsq_cv(A, b, order=1):
    A = np.array(A)
    b = np.array(b)
    m,l = A.shape

    if order==0:
        L = np.eye(l)
    elif order==1:
        L = np.zeros((l-1,l))
        L[:,:-1] -= np.eye(l-1)
        L[:,1:] += np.eye(l-1)
    elif order==2:
        L = np.zeros((l-2,l))
        L[:,:-2] += np.eye(l-2)
        L[:,1:-1] -= 2*np.eye(l-2)
        L[:,2:] += np.eye(l-2)
    elif order is None:
        L = np.zeros(1)
    else:
        L = np.array(order)
        assert L.shape[-1]==l or L.shape in ((), (1,))

    return linalg.lstsq(A, b)



"""
Methods for performing Gaussian quadrature.

Examples
--------
A quadrature rule where the limits are defined by floats::

    >>> abscissas, weights = chaospy.quad_clenshaw_curtis(
    ...         order=3, lower=0., upper=1.)
    >>> print(abscissas)
    [[ 0.    0.25  0.75  1.  ]]
    >>> print(weights)
    [ 0.11111111  0.38888889  0.19444444  0.11111111]

A quadrature rule in higher dimensions through the main interface::

    >>> lower = [0., 0.]
    >>> upper = [1., 1.]
    >>> abscissas, weights = chaospy.generate_quadrature(
    ...         order=1, domain=(lower, upper), rule="gauss_legendre")
    >>> print(abscissas)
    [[ 0.21132487  0.21132487  0.78867513  0.78867513]
     [ 0.21132487  0.78867513  0.21132487  0.78867513]]
    >>> print(weights)
    [ 0.25  0.25  0.25  0.25]

A quadrature rule where the distribution is the weight function::

    >>> dist = chaospy.Gamma()
    >>> abscissas, weights = chaospy.quad_golub_welsch(
    ...         order=2, dist=dist)
    >>> print(abscissas)
    [[ 0.41577456  2.29428036  6.28994508]]
    >>> print(weights)
    [ 0.71109301  0.27851773  0.01038926]
"""
from .combine import combine
from .interface import generate_quadrature, normalize_weights

from .stieltjes import generate_stieltjes
from .collection import *
from .sparse_grid import sparse_grid
from .generator import rule_generator

import chaospy



"""
Antithetic variables.
"""

import numpy as np

def antithetic_gen(U, order):

    order = np.array(order, dtype=bool).flatten()
    if order.size==1 and len(U)>1:
        order = np.repeat(order, len(U))

    U = np.asfarray(U).T
    iU = 1-U
    out = []
    index = np.zeros(len(U.T), dtype=bool)

    def expand(I):
        index[order] = I
        return index

    for I in np.ndindex((2,)*sum(order*1)):
        I = expand(I)
        out.append((U*I + iU*(True-I)).T)

    out = np.concatenate(out[::-1], 1)
    return out



r"""
Constructing an transitional operator requires to construct the
distributions different from a stand-alone variable. To illustrate how
to construct these types of variables, consider the following example.
Let :math:`Q=A+B`, where one of :math:`A` and :math:`B` is a random
variable, and the other a scalar. Which variable is what dependents on
the user setup of the variable. Assuming that :math:`A` is the random
variable, we have that

.. math::

    F_{Q\mid B}(q\mid b) = \mathbb P {q\leq Q\mid B\!=\!b} =
    \mathbb P {q\leq AB\mid B\!=\!b}

    = \mathbb P {\tfrac qb\leq A\mid B\!=\!b} =
    F_{A\mid B}(\tfrac qb\mid b).

Because of symmetry the distribution will be the same, but with
:math:`A` and :math:`B` substituted.

This is required when trying to use operators on multivariate variables. To
create such a variable with ``construct`` provide an additional ``length``
keyword argument specifying the length of a distribution.
"""

import numpy as np

import networkx as nx
from .approx import pdf, ppf, mom


class Graph():
    """
    Class for organizing dependencies in runtime.

    Iteraction with this element is required to create advanced
    distributions.

    To use a let a distribution be advanced, include the following
    parameters on initiation::

        _advance=True       Initiates the advanced distribution
        _length=len(dist)   Te length of a distribution. Defaults to 1

    Subsetting the Dist-class using the _advance feature requires the
    following backend methods.
    The required ones are::

        _cdf(self, x, graph)        Cumulative distribution function
        _bnd(self, x, graph)        Upper and lower bounds at location x

    The following can be provided::

        _pdf(self, x, graph)        Probability density function
        _ppf(self, q, graph)        CDF inverse
        _mom(self, k, graph)        Statistical moment generator
        _ttr(self, k, graph)        TTR coefficients generator
        _dep(self, graph)           Dependency callback.
        _val(self, graph)           Value callback function. Used to
                                itterative check if a distribution has
                                been evaluated before, indirectly.
        _str(self, **prm)       Preaty print of distribution

    Access parameter for distributions, input locations (or
    constants), and function evaluations respectively::

        graph.dists, graph.keys, graph.values

    See Container for respective usage usage.

    Iterate through the graph by evaluating the distribution it is a wrapper
    for::

        graph(z, dist)

    where z is locations, and dist is the distribution it wraps.

    Iterate through distribution's dependencies::

        for dist in graph

    or get all distributions as a list::

        graph.dependencies()

    print the chaospy state of graph::

        print(graph)

    Make a copy of chaospy state::

        graph.copy()

    Initiate an itteration process::

        graph.run(...)

    Switch from PDF to forward seemlessly::

        graph.fwd_as_pdf(x, dist)

    From dist.baseclass we have that there are two new methods. They are
    defined as follows::

        _val(self, graph)   If enough information exists to say that the
                        distribution is evaluated given the state of
                        parameters, return that value.  Else return self.

        _dep(self, graph)   Return a list l with len(l)==len(self) with all
                        distributions it depends upon as a set in each
                        element.  A common element between two sets implies
                        dependencies.  Use graph(dist) (instead of graph(x, dist)) to
                        generate a distributions dependencies.
    """

    def __init__(self, dist):
        """
        Args:
            dist (Dist) : The root of the dependency tree.
        """

        graph = nx.DiGraph()
        L = [dist]
        graph.add_node(dist)

        while L:

            d = L.pop()
            for key,val in d.prm.items():

                if not isinstance(val, np.ndarray):
                    if not (val in graph.node):
                        graph.add_node(val)
                        L.append(val)
                    graph.add_edge(d, val)

        assert nx.is_directed_acyclic_graph(graph)

        self.graph_source = graph

        self.valmode = False
        self.size = None
        self.meta = {}
        self.graph = graph
        self.L = nx.topological_sort(graph)
        self.root = self.dist = dist

        self.dists = Dists(self)
        self.keys = Keys(self)
        self.values = Values(self)

        self._call = None


    def __call__(self, *args, **kwargs):
        return self._call(*args, **kwargs)


    def __str__(self):
        graph = self.graph
        out = ""
        for dist in self.L:
            out += str(dist) + "\n"
            for key in ["key", "val"]:
                if key in graph.node[dist]:
                    out += key + "\n"
                    out += str(graph.node[dist][key]) + "\n"
            for key in ["inv", "fwd", "pdf", "rnd",
                    "range"]:
                if key in graph.node[dist]:
                    out += key + ": "
                    out += str(graph.node[dist][key]) + "\n"
            out += "\n"

        out = out[:-1]
        return out

    def __iter__(self):
        return self.L.__iter__()

    def copy(self):
        """
        Shallow copy of graph. Distribution stays the same.
        """

        graph = Graph(self.root)
        for node in self.graph.nodes():
            graph.graph.add_node(node, **self.graph.node[node])
        graph.graph.add_edges_from(self.graph.edges())

        graph.size = self.size
        graph.meta = self.meta
        graph._call = self._call

        return graph


    def run(self, x, mode, **meta):
        """
        Run through network to perform an operator.

        Using this during another call is possible, but care has to be
        taken to ensure that the chaospy graph state is compatible with the
        new call.

        Available modes to run in:
        ------- -------------------------------------------
        "fwd"   Forward
        "inv"   Inverse
        "range" Lower and upper limits
        "ttr"   Three terms recursion coefficient generator
        "mom"   Raw statistical moments
        "dep"   Dependencies
        "val"   Values
        ------- -------------------------------------------

        Args:
            x (np.ndarray) : Input variable with shape=(D,K), where D is the
                    number of dimensions in dist.
            mode (str) : What type of operator to perform.
        **meta
            Keyword argument passed to approximation method if run.
        """
        if mode in ("rnd", "dep", "val"):
            self.size, size = x, self.size
        else:
            self.size, size = x.shape[-1], self.size
        self.meta, meta = meta, self.meta

        if mode != "val":
            graph_source = self.graph_source
            graph = nx.DiGraph()
            for node in graph_source.nodes():
                graph.add_node(node)
            graph.add_edges_from(graph_source.edges())
            graph, self.graph = self.graph, graph

        call = self._call
        if mode=="fwd":
            self._call = self.fwd_call
        elif mode=="pdf":
            self._call = self.pdf_call
        elif mode=="inv":
            self._call = self.inv_call
        elif mode=="range":
            self._call = self.range_call
        elif mode=="ttr":
            self._call = self.ttr_call
        elif mode=="mom":
            self._call = self.mom_call
        elif mode=="dep":
            self._call = self.dep_call
        elif mode=="rnd":
            self._call = self.rnd_call
        elif mode=="val":
            self._call = self.val_call
        else:
            raise ValueError("unknown mode")

        if mode!="val":
            self.dist = self.root

        if mode in ("rnd","dep"):
            out = self(self.root)

        elif mode=="val":
            out = self(x)
        else:
            out = self(x, self.root)

        if mode=="ttr":
            out[1]**(x!=0)

        self.size = size
        if mode=="val":
            graph = self.graph
        else:
            graph, self.graph = self.graph, graph
        self._call = call
        self.meta = meta

        return out, graph

    def counting(self, dist, mode):
        "counter function. Used for meta analysis."
        if dist in self.graph.node and \
                mode in self.graph.node[dist]:
                    self.graph.node[dist][mode] += 1
        else:
            self.graph.add_node(dist, **{mode:1})

    def fwd_call(self, x, dist):
        "forward call backend wrapper"

        self.counting(dist, "fwd")
        assert x.shape==(len(dist), self.size)
        self.dist, dist_ = dist, self.dist

        graph = self.graph
        graph.add_node(dist, key=x)
        out = np.empty(x.shape)

        prm = self.dists.build()
        prm.update(self.keys.build())
        for k,v in prm.items():
            if not isinstance(v, np.ndarray):
                v_ = self.run(v, "val")[0]
                if isinstance(v_, np.ndarray):
                    prm[k] = v_
                    graph.add_node(v, key=v_)

        if dist.advance:
            out[:] = dist._cdf(x, self)
        else:
#              prm = self.dists.build()
#              prm.update(self.keys.build())
            out[:] = dist._cdf(x, **prm)

        graph.add_node(dist, val=out)

        self.dist = dist_
        return np.array(out)

    def pdf_call(self, x, dist):
        "PDF call backend wrapper"

        self.counting(dist, "pdf")
        assert x.shape==(len(dist), self.size)
        self.dist, dist_ = dist, self.dist

        graph = self.graph
        graph.add_node(dist, key=x)
        out = np.empty(x.shape)

        prm = self.dists.build()
        prm.update(self.keys.build())
        for k,v in prm.items():
            if not isinstance(v, np.ndarray):
                v_ = self.run(v, "val")[0]
                if isinstance(v_, np.ndarray):
                    prm[k] = v_
                    graph.add_node(v, key=v_)

        if hasattr(dist, "_pdf"):
            if dist.advance:
                out[:] = dist._pdf(x, self)
            else:
#                  prm = self.dists.build()
#                  prm.update(self.keys.build())
                out[:] = dist._pdf(x, **prm)
        else:
            out = pdf(dist, x, self, **self.meta)
        graph.add_node(dist, val=out)

        self.dist = dist_
        return np.array(out)

    def fwd_as_pdf(self, x, dist):
        """During a PDF-call, a switch to CDF might be necesarry.
        This functions initiates this switch."""

        graph_source = self.graph_source
        graph = nx.DiGraph()
        for node in graph_source.nodes():
            graph.add_node(node)
        graph.add_edges_from(graph_source.edges())
        graph, self.graph = self.graph, graph

        self._call = self.fwd_call
        out = self(x, dist)

        graph, self.graph = self.graph, graph
        self._call = self.pdf_call

        return out


    def inv_call(self, q, dist):
        "inverse call backend wrapper"

        self.counting(dist, "inv")
        assert q.shape==(len(dist), self.size)
        self.dist, dist_ = dist, self.dist

        graph = self.graph
        graph.add_node(dist, val=q)
        out = np.empty(q.shape)

        prm = self.dists.build()
        prm.update(self.keys.build())
        for k,v in prm.items():
            if not isinstance(v, np.ndarray):
                v_ = self.run(v, "val")[0]
                if isinstance(v_, np.ndarray):
                    prm[k] = v_
                    graph.add_node(v, key=v_)

        if hasattr(dist, "_ppf"):
            if dist.advance:
                out[:] = dist._ppf(q, self)
            else:
                out[:] = dist._ppf(q, **prm)
        else:
            out,N,q_ = ppf(dist, q, self,
                    retall=1, **self.meta)
        graph.add_node(dist, key=out)

        self.dist = dist_
        return np.array(out)

    def range_call(self, x, dist):
        "range call backend wrapper"

        self.counting(dist, "range")
        assert x.shape==(len(dist), self.size)
        graph = self.graph
        self.dist, dist_ = dist, self.dist

        graph.add_node(dist, key=x)
        out = np.empty((2,)+x.shape)

        prm = self.dists.build()
        prm.update(self.keys.build())
        for k,v in prm.items():
            if not isinstance(v, np.ndarray):
                v_ = self.run(v, "val")[0]
                if isinstance(v_, np.ndarray):
                    prm[k] = v_
                    graph.add_node(v, key=v_)

        if dist.advance:
            _ = dist._bnd(x, self)
            out[0,:],out[1,:] = _
        else:
            lo, up = dist._bnd(**prm)
            lo, up = np.array(lo), np.array(up)
            out.T[:,:,0],out.T[:,:,1] = lo.T, up.T
        graph.add_node(dist, val=out)

        self.dist = dist_
        return np.array(out)

    def ttr_call(self, k, dist):
        "TTR call backend wrapper"

        assert k.shape==(len(dist), self.size)
        graph = self.graph
        self.dist, dist_ = dist, self.dist

        graph.add_node(dist, key=k)
        if hasattr(dist, "_ttr"):
            if dist.advance:
                out = dist._ttr(k, self)
            else:
                out = np.empty((2,)+k.shape)
                prm = self.dists.build()
                prm.update(self.keys.build())
                out[0],out[1] = dist._ttr(k, **prm)
        else:
            raise NotImplementedError()

        graph.add_node(dist, val=out)

        self.dist = dist_
        return np.array(out)

    def mom_call(self, k, dist):
        "Moment generator call backend wrapper"
        assert len(k)==len(dist)
        graph = self.graph
        self.dist, dist_ = dist, self.dist

        graph.add_node(dist, key=k)
        if hasattr(dist, "_mom"):
            if dist.advance:
                out = dist._mom(k, self)
            else:
                out = np.empty(k.shape[1:])
                prm = self.dists.build()
                prm.update(self.keys.build())
                out[:] = dist._mom(k, **prm)
        else:
            out = mom(dist, k, **self.meta)
        graph.add_node(dist, val=out)

        self.dist = dist_
        return np.array(out)

    def rnd_call(self, dist):
        "Sample generator call backend wrapper"

        self.counting(dist, "rnd")
        graph = self.graph
        self.dist, dist_ = dist, self.dist

        for k,v in dist.prm.items():
            if not isinstance(v, np.ndarray) and\
                    not "key" in graph.node[v]:
                self.rnd_call(v)

        if dist.advance:
            key, _ = self.run(dist, "val")
        else:
            rnd = np.random.random((len(dist),self.size))
            key = self.inv_call(rnd, dist)

        assert isinstance(key, np.ndarray)
        graph.add_node(dist, key=key)

        self.dist = dist_

        if dist is self.root:
            out = graph.node[dist]["key"]
            return out


    def dep_call(self, dist):
        "Dependency call backend wrapper"

        graph = self.graph
        self.dist, dist_ = dist, self.dist

        if hasattr(dist, "_dep"):
            out = dist._dep(self)
        else:
            for val in self.prm.values():
                out = np.zeros(len(dist), dtype=bool)
                if val in graph.nodes:
                    out = np.ones(len(dist), dtype=bool)
                    break
                elif not isinstance(val, np.ndarray):
                    graph.add_node(val)
        graph.add_node(dist, key=out)

        assert len(out)==len(dist)
        self.dist = dist_
        return out

    def val_call(self, dist):
        "Value callback wrapper"

        graph = self.graph

        if "key" in graph.node[dist]:
            return graph.node[dist]["key"]

        self.dist, dist_ = dist, self.dist
        for k,v in dist.prm.items():
            if not isinstance(v, np.ndarray) and\
                    not "key" in graph.node[v]:
                        self.val_call(v)

        if hasattr(dist, "_val"):
            out = dist._val(self)

            if isinstance(out, np.ndarray):
                graph.add_node(dist, key=out)

        else:
            out = dist

        self.dist = dist_
        return out

    def dependencies(self):
        """
Set of node dependencies
        """
        return set(self.graph.nodes())


class Container(object):
    """
    Graph interactive wrapper.

    Used to retrieve parameters in a distribution. Comes in three flavoers::

        Dist Keys Values

    Can be interacted with as follows:

    Retrieve parameter::

        container["param"]

    Check of paramater is in collection::

        "param" in container

    Itterate over all parameters::

        for (key,param) in container

    Generate dictionary with all values::

        container.build()

    Identify the number of parameters available::

        len(container)
    """

    def __init__(self, graph):
        "Graph module"
        self.graph = graph
    def __contains__(self, key):
        raise NotImplementedError()
    def getitem(self, key):
        raise NotImplementedError()
    def __getitem__(self, key):
        return self.getitem(key)
    def build(self):
        "build a dict with all parameters"
        out = {}
        for k,v in self.graph.dist.prm.items():
            if k in self:
                out[k] = self[k]
        return out
    def __iter__(self):
        return self.build().values().__iter__()
    def __str__(self):
        return str(self.build())
    def __len__(self):
        return len(self.build())

class Dists(Container):
    """Contains all distributions."""

    def __contains__(self, key):
        return not isinstance(self.graph.dist.prm[key], np.ndarray)

    def getitem(self, key):
        if not key in self:
            raise KeyError()
        return self.graph.dist.prm[key]

class Keys(Container):
    """
    Contains all values.

    Either as constants or as substitutes of distributions.
    """

    def __contains__(self, key):

        out = False
        val = self.graph.dist.prm[key]
        if isinstance(val, np.ndarray) or \
                "key" in self.graph.graph.node[val]:
            out = True

        return out

    def getitem(self, key):

        if not key in self:
            raise KeyError()
        val = self.graph.dist.prm[key]

        if isinstance(val, np.ndarray):
            return val

        gkey = self.graph.dist.prm[key]
        return self.graph.graph.node[gkey]["key"]

class Values(Container):
    """Contains all evaluations of distributions."""

    def __contains__(self, key):
        out = self.graph.dist.prm[key]
        return  not isinstance(out, np.ndarray) and \
                "val" in self.graph.graph.node[out]

    def getitem(self, key):
        if not key in self:
            raise KeyError()
        return self.graph.graph.node[self.graph.dist.prm[key]]["val"]



r"""
As noted throughout the documentation, known distributions are created easily
by calling their name of the function of interest. For example to create
a Gaussian random variable::

   >>> distribution = cp.Normal(0,1)

To construct simple multivariate random variables with stochastically
independent components, either all the same using :func:`~chaospy.dist.Iid`::

   >>> distribution = cp.Iid(cp.Normal(0, 1), 2)

Or with more detailed control through :func:`~chaospy.dist.J`::

   >>> distribution = cp.J(cp.Normal(0, 1), cp.Normal(0, 1))

The functionality of the distributions are covered in various other sections:

* To generate random samples, see :ref:`montecarlo`.
* To create transformations, see :ref:`rosenblatt`.
* To generate raw statistical moments, see :ref:`moments`.
* To generate three terms recurrence coefficients, see :ref:`orthogonality`.
* To analyse statistical properies, see :ref:`descriptives`.

"""

import chaospy.dist.baseclass
from chaospy.dist.baseclass import Dist

import chaospy.dist.graph
import chaospy.dist.sampler
import chaospy.dist.approx
import chaospy.dist.joint
import chaospy.dist.cores
import chaospy.dist.copulas
import chaospy.dist.collection
import chaospy.dist.operators
import chaospy.dist.rosenblatt

from chaospy.dist.graph import *
from chaospy.dist.sampler import *
from chaospy.dist.approx import *
from chaospy.dist.joint import *
from chaospy.dist.cores import *
from chaospy.dist.copulas import *
from chaospy.dist.collection import *
from chaospy.dist.operators import *

from numpy.random import seed


if __name__ == "__main__":
    seed(1000)
    import doctest
    import chaospy as cp
    doctest.testmod()



"""
Generates quasirandom Sobol vector.

This code is distributed under the GNU LGPL license.

The routine adapts the ideas of Antonov and Saleev.
Original FORTRAN77 version by Bennett Fox.
MATLAB version by John Burkardt.
PYTHON version by Corrado Chisari

Reference:

      Antonov, Saleev,
      USSR Computational Mathematics and Mathematical Physics,
      Volume 19, 1980, pages 252 - 256.

      Paul Bratley, Bennett Fox,
      Algorithm 659:
      Implementing Sobol's Quasirandom Sequence Generator,
      ACM Transactions on Mathematical Software,
      Volume 14, Number 1, pages 88-100, 1988.

      Bennett Fox,
      Algorithm 647:
      Implementation and Relative Efficiency of Quasirandom 
      Sequence Generators,
      ACM Transactions on Mathematical Software,
      Volume 12, Number 4, pages 362-376, 1986.

      Ilya Sobol,
      USSR Computational Mathematics and Mathematical Physics,
      Volume 16, pages 236-242, 1977.

      Ilya Sobol, Levitan, 
      The Production of Points Uniformly Distributed in a Multidimensional 
      Cube (in Russian),
      Preprint IPM Akad. Nauk SSSR, 
      Number 40, Moscow 1976.
"""

import numpy as np
__author__ = "Jonathan Feinberg"
__email__ = "jonathan@feinberg.no"
__date__ = "2013-05-28"

_seed = 1

def bit_hi(n):
    bit = 0
    while True:
        if n <= 0:
            break
        bit += 1
        n = int(n/2)
    return bit

def bit_lo(n):
    bit = 0
    while True:
        bit += 1
        i = int(n/2)
        if n==2*i: break
        n = i
    return bit


def sobol(dim_num, N, seed=None):
    """
Parameters
----------
dim_num : int
    Number of spacial dimensions.
    Must satisfy 0<dim_num<41
N : int
    Number of unique samples to generate
seed : int, optional
    Starting seed. Non-positive values are treated as 1.
    If omited, consequtive samples are used.

Returns
-------
quasi : np.ndarray
    Quasi-random vector with `shape=(dim_num, N)`.
    """

    global _seed

    dim_max = 40
    log_max = 30

    if seed is None:
        seed = _seed
    _seed += N

    v = np.zeros((dim_max,log_max), dtype=int)
    v[0:40,0] = 1

    v[2:40,1] = 1, 3, 1, 3, 1, 3, 3, 1, \
                3, 1, 3, 1, 3, 1, 1, 3, 1, 3, \
                1, 3, 1, 3, 3, 1, 3, 1, 3, 1, \
                3, 1, 1, 3, 1, 3, 1, 3, 1, 3

    v[3:40,2] = 7, 5, 1, 3, 3, 7, 5, \
                5, 7, 7, 1, 3, 3, 7, 5, 1, 1, \
                5, 3, 3, 1, 7, 5, 1, 3, 3, 7, \
                5, 1, 1, 5, 7, 7, 5, 1, 3, 3

    v[5:40,3] = 1, 7, 9,13,11, \
                1, 3, 7, 9, 5,13,13,11, 3,15, \
                5, 3,15, 7, 9,13, 9, 1,11, 7, \
                5,15, 1,15,11, 5, 3, 1, 7, 9

    v[7:40,4] = 9, 3,27, \
               15,29,21,23,19,11,25, 7,13,17, \
                1,25,29, 3,31,11, 5,23,27,19, \
               21, 5, 1,17,13, 7,15, 9,31, 9

    v[13:40,5] = 37,33, 7, 5,11,39,63, \
                 27,17,15,23,29, 3,21,13,31,25, \
                  9,49,33,19,29,11,19,27,15,25

    v[19:40,6] = 13, \
                 33,115, 41, 79, 17, 29,119, 75, 73,105, \
                  7, 59, 65, 21,  3,113, 61, 89, 45,107

    v[37:40,7] = 7, 23, 39

    poly=    1,   3,   7,  11,  13,  19,  25,  37,  59,  47, \
            61,  55,  41,  67,  97,  91, 109, 103, 115, 131, \
           193, 137, 145, 143, 241, 157, 185, 167, 229, 171, \
           213, 191, 253, 203, 211, 239, 247, 285, 369, 299

    atmost = 2**log_max - 1

    #Find the number of bits in ATMOST.
    maxcol = bit_hi ( atmost )

    #Initialize row 1 of V.
    v[0,0:maxcol] = 1

    #Check parameters.
    assert 0 < dim_num < dim_max, "dim_num in [1,40]"

    #Initialize the remaining rows of V.
    for i in range(1 , dim_num):

        #The bits of the integer POLY(I) gives the form of polynomial I.
        #Find the degree of polynomial I from binary encoding.
        j = poly[i]
        m = 0
        while True:
            j = int(j/2)
            if j <= 0:
                break
            m += 1

        #Expand this bit pattern to separate components of the logical array INCLUD.
        j = poly[i]
        includ = np.zeros(m, dtype=bool)
        for k in range(m, 0, -1):
            j2 = int(j/2)
            includ[k-1] =  j != 2*j2
            j = j2

        #Calculate the remaining elements of row I as explained
        #in Bratley and Fox, section 2.
        for j in range(m+1, maxcol+1):
            newv = v[i, j-m-1].item()
            l = 1
            for k in range(1, m+1):
                l = 2 * l
                if includ[k-1]:
                    newv = newv ^ l*v[i,j-k-1].item()

            v[i,j-1] = newv

    v = v[:dim_num]


    #Multiply columns of V by appropriate power of 2.
    l = 1
    for j in range(maxcol-1, -1, -1):
        l = 2 * l
        v[:,j] = v[:,j] * l

    #RECIPD is 1/(common denominator of the elements in V).
    recipd = 1. / ( 2 * l )
    lastq=np.zeros(dim_num, dtype=int)

    seed = int(seed)
    if seed < 1:
        seed = 1

    for seed_ in range(seed):
        lastq[:] = lastq ^ v[:,bit_lo(seed_)-1]

    #Calculate the new components of QUASI.
    quasi=np.empty((dim_num, N))
    for j in range(N):
        quasi[:, j] = lastq * recipd
        lastq[:] = lastq ^ v[:,bit_lo(seed+j)-1]

    return quasi




"""
Monte Carlo simulation is by nature a very slow converging method.  The error
in convergence is proportional to :math:`1/\sqrt{K}` where :math:`K` is the
number of samples.  It is somewhat better with variance reduction techniques
that often reaches errors proportional to :math:`1/K`. For a
full overview of the convergence rate of the various methods, see for
example :cite:`kroese_handbook_2011`. However as the number
of dimensions grows, Monte Carlo convergence rate stays the same, making it
immune to the curse of dimensionality.

Generating random samples can be done from the distribution instance
method ``sample`` as discussed in the :ref:`tutorial`.
For example, to generate nodes from the Korobov latice::

   >>> distribution = pc.Iid(pc.Beta(2, 2), 2)
   >>> samples = distribution.sample(4, rule="K")
   >>> print(nodes)
   [[ 0.28714073  0.43293108  0.56706892  0.71285927]
    [ 0.43293108  0.71285927  0.28714073  0.56706892]]


At the core of `chaospy`, all samples are generated to the unit hyper-cube
through the function :func:`~chaospy.dist.samplegen`. These samples are then
mapped using a :ref:`rosenblatt` to map the values on the domain respective to
the distribution in question. This way, all variance reduction techniques are
supported by all distributions.
"""

import numpy as np

from .generator import samplegen
from .samplers import *
from .sobol_lib import sobol
from .antithetic import antithetic_gen


if __name__=="__main__":
    import __init__ as cp
    import doctest
    doctest.testmod()



"""
Functions for generating samples on the unite hyper cube.
"""

import numpy as np

import chaospy.quad

def chebyshev(dim, n):
    "Chebyshev sampling"
    x = .5*np.cos(np.arange(n,0,-1)*np.pi/(n+1)) + .5
    X = chaospy.quad.combine([x]*dim)
    return X.T


def chebyshev_nested(dim, n):
    "Chebyshev sampling, nested"
    x = .5*np.cos(np.arange(2**n-1,0,-1)*np.pi/(2**n)) + .5
    X = chaospy.quad.combine([x]*dim)
    return X.T


def regular_grid(dim, n):
    "regular grid"
    x = np.arange(1,n+1)/(n+1.)
    X = chaospy.quad.combine([x]*dim)
    return X.T


def regular_grid_nested(dim, n):
    "regular grid, nested"
    x = np.arange(1,2**n)*1./(2**n)
    X = chaospy.quad.combine([x]*dim)
    return X.T

def corput(p, k):
    """
Van der Corput sampling
    """
    k += 1
    out = 0.
    p_base = p
    while k > 0:
        a = k % p_base
        out += a*1./p
        k = int(k / p_base)
        p *= p_base
    return out


def primes(n):
    """
    Generate primes using sieve of Eratosthenes.

    Examples:
        >>> print(primes(20))
        [2, 3, 5, 7, 11, 13, 17, 19]
    """
    if n == 2:
        return [2]

    elif n < 2:
        return []

    s = list(range(3, n+1, 2))
    mroot = n ** 0.5
    half = int((n+1)/2-1)
    i = 0
    m = 3
    while m <= mroot:
        if s[i]:
            j = int((m*m-3)/2)
            s[j] = 0
            while j < half:
                s[j] = 0
                j += m
        i = i+1
        m = 2*i+3
    out = [2] + [x for x in s if x]
    return out


def hammersley(dim, n):
    """
    Hammersley sequence.

    Examples:
        >>> print(hammersley(2, 4))
        [[ 0.2    0.4    0.6    0.8  ]
         [ 0.5    0.25   0.75   0.125]]
    """
    p = []
    m = 10
    while len(p)<dim:
        p = primes(m)
        m *= 2
    p = p[:dim]

    out = np.zeros((dim, n))
    out[0] = np.arange(1, n+1)*1./(n+1)
    for i in range(1, dim):
        for j in range(n):
            out[i,j] = corput(p[i-1], j)
    return out


def halton(dim, n):
    """
    Halton sequence.

    Examples:
        >>> print(halton(2, 4))
        [[ 0.125       0.625       0.375       0.875     ]
         [ 0.44444444  0.77777778  0.22222222  0.55555556]]
    """
    p = []
    m = 10
    while len(p) < dim:
        p = primes(m)
        m *= 2

    p = p[:dim]
    burn = p[-1]

    out = np.empty((n, dim))
    for i in range(n):
        for j in range(dim):
            out[i,j] = corput(p[j], i+burn)
    return out.T


def korobov(dim, n, a=17797):
    """
    Korobov set.

    Examples:
        >>> print(korobov(2, 4))
        [[ 0.2  0.4  0.6  0.8]
         [ 0.4  0.8  0.2  0.6]]
    """
    z = np.empty(dim)
    z[0] = 1
    for i in range(1,dim):
        z[i] = a*z[i-1] % (n+1)

    grid = np.mgrid[:dim,:n+1]
    Z = z[grid[0]]
    B = grid[1]+1
    P = B*Z/(n+1) % 1
    return P[:,:n]


def latin_hypercube(dim, n):
    """
    Latin Hypercube sampling.

    Examples:
        >>> cp.seed(1000)
        >>> print(cp.latin_hypercube(2, 4))
        [[ 0.6633974   0.27875174  0.98757072  0.12054785]
         [ 0.46811863  0.05308317  0.51017741  0.84929862]]
    """
    R = np.random.random(n*dim).reshape((dim, n))
    for d in range(dim):
        perm = np.random.permutation(n)
        R[d] = (perm + R[d])/n
    return R

if __name__ == "__main__":
    import chaospy as cp
    import doctest
    doctest.testmod()



"""
The Saltelli method

Code is built upon the code provided by Vinzenze Eck
"""
import numpy as np

class Saltelli:

    def __init__(self, dist, samples, poly=None, rule="R"):
        self.dist = dist
        samples_ = dist.sample(2*samples, rule=rule)
        self.samples1 = samples_.T[:samples].T
        self.samples2 = samples_.T[samples:].T
        self.poly = poly

    def __getitem__(self, indices):
        assert len(self.dist) == len(indices)

        key = "".join([i and "a" or "b" for i in indices])

        if hasattr(self, key):
            return getattr(self, key)

        new = np.empty(self.samples1.shape)
        for i in range(len(indices)):
            if indices[i]:
                new[i] = self.samples1[i]
            else:
                new[i] = self.samples2[i]

        if self.poly:
            new = self.poly(*new)

        setattr(self, key, new)
        return new


def Sens_m_sample(poly, dist, samples, rule="R"):

    dim = len(dist)

    Y = Saltelli(dist, samples, poly)

    ones = [1]*dim
    zeros = [0]*dim
    index = [0]*dim

    V = np.var(Y[zeros], -1)

    yo = Y[ones]
    yz = Y[zeros]
    mean = .5*(np.mean(yo) + np.mean(yz))

    yo -= mean
    yz -= mean

    out = []
    for d in range(dim):

        index[d] = 1

        yi = Y[index]-mean

        s = np.mean(yo*(yi-yz), -1) / (V+(V==0))*(V!=0)
        out.append(s)
        index[d] = 0

    return np.array(out)


def Sens_m2_sample(poly, dist, samples, rule="R"):

    dim = len(dist)

    Y = Saltelli(dist, samples, poly)

    ones = [1]*dim
    zeros = [0]*dim
    index = [0]*dim

    V = np.var(Y[zeros], -1)

    yo = Y[ones]
    yz = Y[zeros]
    mean = .5*(np.mean(yo) + np.mean(yz))

    yo -= mean
    yz -= mean

    out = np.empty((dim,dim)+poly.shape)
    for d1 in range(dim):

        index[d1] = 1
        yi = Y[index]-mean
        s = np.mean(yo*(yi-yz), -1) / (V+(V==0))*(V!=0)
        out[d1,d1] = s

        for d2 in range(d1+1, dim):

            index[d2] = 1

            yi = Y[index]-mean

            s = np.mean(yo*(yi-yz), -1) / (V+(V==0))*(V!=0)
            out[d1,d2] = out[d2,d1] = s

            index[d2] = 0

        index[d1] = 0

    return out


def Sens_t_sample(poly, dist, samples, rule="R"):

    dim = len(dist)
    Y = Saltelli(dist, samples, poly)

    zeros = [0]*dim
    index = [1]*dim

    V = np.var(Y[zeros], -1)

    out = []
    for d in range(dim):

        index[d] = 0
        s = 1-np.mean((Y[index]-Y[zeros])**2, -1) / (2*V+(V==0))*(V!=0)
        out.append(s)
        index[d] = 1

    return np.array(out)

if __name__ == "__main__":

    import __init__ as cp

    rho = 0.5
    poly = cp.basis(0,2,2)
    dist = cp.Iid(cp.Normal(), 2)
    dist_dep = cp.MvNormal([0,0], [[1, rho], [rho, 1]])

    print(poly)
    print("md\n", Sens_m_sample(poly, dist_dep, 10**6))
    print("m2d\n", Sens_m2_sample(poly, dist_dep, 10**6)[0,1])
    print("td\n", Sens_t_sample(poly, dist_dep, 10**6))
    print([[0,1,0,1,1/(1+rho*rho),0], [0,0,1,0,1/(1+rho*rho),1]])



r"""
Copulas are a type dependency structure imposed on independent variables to
achieve to more complex problems without adding too much complexity.

To construct a copula one needs a copula transformation and the
Copula wrapper::

    >>> dist = cp.Iid(cp.Uniform(), 2)
    >>> copula = cp.Gumbel(dist, theta=1.5)

The resulting copula is then ready for use::

    >>> cp.seed(1000)
    >>> print(copula.sample(5))
    [[ 0.65358959  0.11500694  0.95028286  0.4821914   0.87247454]
    [ 0.02388273  0.10004972  0.00127477  0.10572619  0.4510529 ]]
"""

import chaospy.dist.copulas.baseclass
import chaospy.dist.copulas.collection

from chaospy.dist.copulas.baseclass import Copula, Archimedean
from chaospy.dist.copulas.collection import (
    Gumbel, Clayton, Ali_mikhail_haq, Frank, Joe, Nataf, T_copula
)


if __name__=="__main__":
    import chaospy as cp
    import doctest
    doctest.testmod()



"""
Division.
"""
import numpy

from chaospy.dist import Dist
import chaospy.dist.operators

class Div(Dist):
    """Division."""

    def __init__(self, left, right):
        """
        Constructor.

        Args:
            left (Dist, array_like) : Left hand side.
            right (Dist, array_like) : Right hand side.
        """
        left_ = not isinstance(left, Dist) or 1 and len(left)
        right_ = not isinstance(right, Dist) or 1 and len(right)
        if isinstance(left, Dist):
            assert not numpy.any(numpy.prod(left.range(), 0) < 0)
        if isinstance(right, Dist):
            assert not numpy.any(numpy.prod(right.range(), 0) < 0)
        length = max(left_,right_)
        Dist.__init__(self, left=left, right=right,
                _length=length, _advance=True)

    def _pdf(self, xloc, graph):
        """Probability density function."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
            num = (num.T*numpy.ones(xloc.shape[::-1])).T

            p0 = graph.fwd_as_pdf(xloc*0, dist)
            xloc, num = numpy.where(xloc, xloc, 1), \
                    numpy.where(xloc, num, numpy.inf*numpy.sign(num))
            p1 = graph(num/xloc, dist)
            out = numpy.abs((1-2*p0)*p1*num/xloc**2)
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
            num = (num.T*numpy.ones(xloc.T.shape)).T
            num = numpy.where(num, num, numpy.inf)
            out = numpy.abs(graph(xloc*1./num, dist)/num)
        return out

    def _cdf(self, xloc, graph):
        """Cumulative distribution function."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
            num = (num.T*numpy.ones(xloc.shape[::-1])).T
            p0 = graph.copy()(xloc*0, dist)
            xloc, num = numpy.where(xloc, xloc, 1), \
                    numpy.where(xloc, num, numpy.inf*numpy.sign(num))
            p1 = graph(num/xloc, dist)
            out = (1-p0)*(1-p1) + p0*p1
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
            xloc = (xloc.T/(num.T + 1.*(num.T==0))).T
            out = graph(xloc, dist)
            out = (out.T*(num.T>0) + (1.-out.T)*(num.T<0) + \
                    1.*(xloc.T>num.T)*(num.T==0)).T
        return out

    def _ppf(self, q, graph):
        """Point percentile function."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
            num = (num.T*numpy.ones(q.shape[::-1])).T
            q = (q*(num>0) + (1.-q)*(num<=0))
        return graph(q, dist)*num

    def _bnd(self, xloc, graph):
        """Distribution bounds."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]

        num = (num + 1.*(num==0))
        xloc = (xloc.T/num.T).T
        lo_,up_ = graph(xloc, dist)
        lower = (num.T*(lo_.T*(num.T>0) + up_.T*(num.T<0))).T
        upper = (num.T*(up_.T*(num.T>0) + lo_.T*(num.T<0))).T

        return lower,upper

    def _mom(self, k, graph):
        """Statistical moments."""
        if len(graph.dists)==2 and graph.dists["left"].dependent(graph.dists["right"]):
            raise NotImplementedError("dependency")

        left = []
        if "left" in graph.dists:  left.append(graph(k, graph.dists["left"]))
        else:           left.append((graph.keys["left"].T**k.T).T)
        if "right" in graph.dists:  left.append(graph(k, graph.dists["right"]))
        else:           left.append((graph.keys["right"].T**k.T).T)

        return numpy.prod(left[0]*left[1], 0)


    def _ttr(self, n, graph):
        """Three terms recursion coefficients."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]

        a,b = graph(n, dist)
        return a*num, b*num*num


    def _val(self, graph):
        """Value extraction."""
        if len(graph.keys)==2:
            return graph.keys["left"]*graph.keys["right"]
        return self

    def _str(self, left, right):
        """String representation."""
        return str(left) + "*" + str(right)


def div(left, right):
    """
    Distribution division.

    Args:
        left (Dist, array_like) : left hand side.
        right (Dist, array_like) : right hand side.
    """
    if left is right:
        return 1.

    if isinstance(left, Dist):
        if not isinstance(right, Dist):
            return chaospy.dist.operators.mul(left, 1./right)

    elif not isinstance(right, Dist):
        return left/right

    return Div(left, right)



"""
Uncertainty Quantifican Toolbox
===============================

This module contains tools for performing uncertainty
quantification of models.
"""

import chaospy.bertran
import chaospy.chol
import chaospy.descriptives
import chaospy.dist
import chaospy.orthogonal
import chaospy.poly
import chaospy.spectral
import chaospy.quad
import chaospy.saltelli
import chaospy.regression

from chaospy.dist import *
from chaospy.orthogonal import *
from chaospy.poly import *
from chaospy.spectral import *
from chaospy.quad import *
from chaospy.saltelli import *
from chaospy.descriptives import *
from chaospy.regression import *



"""
Constructing custom probability distributions is done in one of two ways:
Sublcassing the :class:`~chaospy.dist.Dist` or by calling
:func:`~chaospy.dist.construct`. They work about the same except for one
methods are defined, while in the other, functions.

A simple example for constructing a simple uniform distribution::

    >>> def cdf(self, x, lo, up):
    ...     return (x-lo)/(up-lo)
    >>> def bnd(self, lo, up):
    ...     return lo, up
    >>> Uniform = cp.construct(cdf=cdf, bnd=bnd)
    >>> dist = Uniform(-3, 3)
    >>> print(dist.fwd([-3, 0, 3]))
    [ 0.   0.5  1. ]

Here ``cdf`` is the dependent cumulative distribution function as defined in
equation , ``bnd`` is a function returning the lower and upper bounds, and
``a`` and ``b`` are distribution parameters.  They take either other
components, or as illustrated: constants.


In addition to ``cdf`` and ``bnd`` there are a few optional arguments. For
example a fully featured uniform random variable is defined as follows::

    >>> def pdf(self, x, lo, up):
    ...     return 1./(up-lo)
    >>> def ppf(self, q, lo, up):
    ...     return q*(up-lo) + lo
    >>> Uniform = cp.construct(
    ...     cdf=cdf, bnd=bnd, pdf=pdf, ppf=ppf)
    >>> dist = Uniform(-3, 3)

There ``pdf`` is probability distribution function and ``ppf`` if the point
percentile function.  These are methods that provides needed functionality for
probabilistic collocation. If they are not provided during construct, they are
estimated as far as possible.

Equivalently constructing the same distribution using subclass:
:func:`~chaospy.dist.construct`::

    >>> class Uniform(cp.Dist):
    ...     def __init__(self, lo=0, up=1):
    ...         cp.Dist.__init__(self, lo=lo, up=up)
    ...     def _cdf(self, x, lo, up):
    ...         return (x-lo)/(up-lo)
    ...     def _bnd(self, lo, up):
    ...         return lo, up
    ...     def _pdf(self, x, lo, up):
    ...         return 1./(up-lo)
    ...     def _ppf(self, q, lo, up):
    ...         return q*(up-lo) + lo
    ...     def _str(self, lo, up):
    ...         return "u(%s%s)" % (lo, up)
    >>> dist = Uniform(-3, 3)
    >>> print(dist.fwd([-3, 0, 3])) # Forward Rosenblatt transformation
    [ 0.   0.5  1. ]

"""
import types
import numpy as np

import chaospy.dist


class Dist(object):
    """
    The distribution backend class.

    Subclass this module to construct a custom distribuiton.

    If direct subclass of Dist, two method must be provided:

    * Cumulative distribution function (CDF): ``_cdf(self, x, **prm)``.
    * Upper and lower bounds ``_bnd(self, **prm)``.

    The following can be provided:

    * Probability density function: ``_pdf(self, x, **prm)``.
    * CDF inverse: ``_ppf(self, q, **prm)``.
    * Statistical moment generator: ``_mom(self, k, **prm)``.
    * TTR coefficients generator: ``_ttr(self, k, **prm)``.
    * Pretty print of distribution: ``_str(self, **prm)``.

    Alternative use the construct generator :func:`~chaospy.dist.construct`.
    """

    __array_priority__ = 9000

    def __init__(self, **prm):
        """
        Args:
            _length (int) : Length of the distribution
            _advanced (bool) : If True, activate advanced mode
            **prm (array_like) : Other optional parameters. Will be assumed when
                    calling any sub-functions.
        """
        for key, val in prm.items():
            if not isinstance(val, Dist):
                prm[key] = np.array(val)

        self.length = int(prm.pop("_length", 1))
        self.advance = prm.pop("_advance", False)
        self.prm = prm.copy()
        self.graph = chaospy.dist.graph.Graph(self)
        self.dependencies = self.graph.run(self.length, "dep")[0]

    def range(self, x=None, retall=False, verbose=False):
        """
        Generate the upper and lower bounds of a distribution.

        Args:
            x (array_like, optional) : The bounds might vary over the sample
                    space. By providing x you can specify where in the space
                    the bound should be taken.  If omited, a (pseudo-)random
                    sample is used.

        Returns:
            (np.ndarray) : The lower (out[0]) and upper (out[1]) bound where
                    out.shape=(2,)+x.shape
        """
        dim = len(self)
        if x is None:
            x = chaospy.dist.approx.find_interior_point(self)
        else:
            x = np.array(x)
        shape = x.shape
        size = int(x.size/dim)
        x = x.reshape(dim, size)

        out, graph = self.graph.run(x, "range")
        out = out.reshape((2,)+shape)

        if verbose>1:
            print(graph)

        if retall:
            return out, graph
        return out

    def fwd(self, x):
        """
        Forward Rosenblatt transformation.

        Args:
            x (array_like) : Location for the distribution function. x.shape
                    must be compatible with distribution shape.

        Returns:
            (ndarray) : Evaluated distribution function values, where
                    out.shape==x.shape.
        """
        return chaospy.dist.rosenblatt.fwd(self, x)

    def cdf(self, x):
        """
        Cumulative distribution function.

        Note that chaospy only supports cumulative distribution funcitons in
        one dimensions.

        Args:
            x (array_like) : Location for the distribution function. x.shape
                    must be compatible with distribution shape.

        Returns:
            (ndarray) : Evaluated distribution function values, where
                    out.shape==x.shape.

        Except:
            (NotImplementedError) : for distributions with dependent
                components.
        """
        if self.dependent():
            raise NotImplementedError("""\
Cumulative distribution function is only available for stocastically \
independent variables""")
        out = chaospy.dist.rosenblatt.fwd(self, x)
        if len(self) > 1:
            out = np.prod(out, 0)
        return out


    def inv(self, q, maxiter=100, tol=1e-5, verbose=False, **kws):
        """
        Inverse Rosenblatt transformation.

        Args:
            q (array_like) : Probabilities to be inverse. If any values are
                    outside [0,1], error will be raised. q.shape must be
                    compatible with diistribution shape.

        Kwargs:
            maxiter (int) : Maximum number of iterations
            tol (float) : Tolerence level

        Returns:
            (ndarray) : Inverted probability values where out.shape==q.shape.
        """
        return chaospy.dist.rosenblatt.inv(self, q, maxiter, tol, **kws)

    def pdf(self, x, step=1e-7, verbose=0):
        """
        Probability density function.

        Args:
            x (array_like) : Location for the density function. x.shape must
                    be compatible with distribution shape.
            step (float, array_like) : The step length given aproximation is
                    used. If array provided, elements are used along each
                    axis.

        Returns:
            (ndarray) : Evaluated density function values. Shapes are related
                    through the identity x.shape=dist.shape+out.shape
        """
        dim = len(self)
        x = np.array(x)
        shape = x.shape
        size = int(x.size/dim)
        x = x.reshape(dim, size)
        out = np.zeros((dim, size))

        (lo, up), graph = self.graph.run(x, "range")
        valids = np.prod((x.T >= lo.T)*(x.T <= up.T), 1, dtype=bool)
        x[:, True-valids] = (.5*(up+lo))[:, True-valids]
        out = np.zeros((dim,size))

        try:
            tmp,graph = self.graph.run(x, "pdf",
                    eps=step)
            out[:,valids] = tmp[:,valids]
        except NotImplementedError:
            tmp,graph = chaospy.dist.approx.pdf_full(self, x, step, retall=True)
            out[:,valids] = tmp[:,valids]
            if verbose:
                print("approx %s.pdf")
        except IndexError:
            pass

        if verbose>1:
            print(self.graph)

        out = out.reshape(shape)
        if dim>1:
            out = np.prod(out, 0)
        return out

    def sample(self, size=(), rule="R", antithetic=None,
            verbose=False, **kws):
        """
        Create pseudo-random generated samples.

        Args:
            size (int,array_like) : The size of the samples to generate.
            rule (str) : Alternative sampling techniques. See
                    :func:`~chaospy.dist.samplegen`.
            antithetic (bool, array_like) : If provided, will be used to setup
                    antithetic variables.  If array, defines the axes to mirror.

        Returns:
            (ndarray) : Random samples with shape (len(self),)+self.shape
        """
        size_ = np.prod(size, dtype=int)
        dim = len(self)
        if dim>1:
            if isinstance(size, (tuple,list,np.ndarray)):
                shape = (dim,) + tuple(size)
            else:
                shape = (dim, size)
        else:
            shape = size

        out = chaospy.dist.sampler.samplegen(size_, self, rule, antithetic)
        try:
            out = out.reshape(shape)
        except:
            if len(self)==1:
                out = out.flatten()
            else:
                out = out.reshape(dim, int(out.size/dim))

        return out

    def mom(self, K, **kws):
        """
        Raw statistical moments.

        Creates non-centralized raw moments from the random variable. If
        analytical options can not be utilized, Monte Carlo integration
        will be used.

        Args:
            K (array_like) : Index of the raw moments. k.shape must be
                    compatible with distribution shape.  Sampling scheme when
                    performing Monte Carlo
            rule (str) : rule for estimating the moment if the analytical
                    method fails.
            composit (int, array_like optional) : If provided, composit
                    quadrature will be used.  Ignored in the case if
                    gaussian=True.  If int provided, determines number of even
                    domain splits. If array of ints, determines number of even
                    domain splits along each axis. If array of arrays/floats,
                    determines location of splits.
            antithetic (array_like, optional) : List of bool. Represents the
                    axes to mirror using antithetic variable during MCI.

        Returns:
            (ndarray) : Shapes are related through the identity
                    `k.shape==dist.shape+k.shape`.
        """
        K = np.array(K, dtype=int)
        shape = K.shape
        dim = len(self)

        if dim > 1:
            shape = shape[1:]

        size = int(K.size/dim)
        K = K.reshape(dim, size)

        try:
            out, graph = self.graph.run(K, "mom", **kws)

        except NotImplementedError:
            out = chaospy.dist.approx.mom(self, K, **kws)

        return out.reshape(shape)

    def ttr(self, k, acc=10**3, verbose=1):
        """
        Three terms relation's coefficient generator

        Args:
            k (array_like, int) : The order of the coefficients.
            acc (int) : Accuracy of discretized Stieltjes if analytical
                    methods are unavailable.

        Returns:
            (Recurrence coefficients) : Where out[0] is the first (A) and
                    out[1] is the second coefficient With
                    `out.shape==(2,)+k.shape`.
        """
        k = np.array(k, dtype=int)
        dim = len(self)
        shape = k.shape
        shape = (2,) + shape
        size = int(k.size/dim)
        k = k.reshape(dim, size)

        out, graph = self.graph.run(k, "ttr")
        return out.reshape(shape)

    def _ttr(self, *args, **kws):
        """Default TTR generator, throws error."""
        raise NotImplementedError

    def _dep(self, graph):
        """
        Default dependency module backend.

        See graph for advanced distributions.
        """
        sets = [graph(dist) for dist in graph.dists]
        if len(self)==1:
            out = [set([self])]
        else:
            out = [set([]) for _ in range(len(self))]
        for s in sets:
            for i in range(len(self)):
                out[i].update(s[i])

        return out


    def __str__(self):
        """X.__str__() <==> str(X)"""
        if hasattr(self, "_str"):
            return str(self._str(**self.prm))
        return "D"

    def __len__(self):
        """X.__len__() <==> len(X)"""
        return self.length

    def __add__(self, X):
        """Y.__add__(X) <==> X+Y"""
        return chaospy.dist.operators.add(self, X)

    def __radd__(self, X):
        """Y.__radd__(X) <==> Y+X"""
        return chaospy.dist.operators.add(self, X)

    def __sub__(self, X):
        """Y.__sub__(X) <==> X-Y"""
        return chaospy.dist.operators.add(self, -X)

    def __rsub__(self, X):
        """Y.__rsub__(X) <==> Y-X"""
        return chaospy.dist.operators.add(X, -self)

    def __neg__(self):
        """X.__neg__() <==> -X"""
        return chaospy.dist.operators.neg(self)

    def __mul__(self, X):
        """Y.__mul__(X) <==> X*Y"""
        return chaospy.dist.operators.mul(self, X)

    def __rmul__(self, X):
        """Y.__rmul__(X) <==> Y*X"""
        return chaospy.dist.operators.mul(self, X)

    def __div__(self, X):
        """Y.__div__(X) <==> Y/X"""
        return chaospy.dist.operators.mul(self, X**-1)

    def __rdiv__(self, X):
        """Y.__rdiv__(X) <==> X/Y"""
        return chaospy.dist.operators.mul(X, self**-1)

    def __floordiv__(self, X):
        """Y.__floordiv__(X) <==> Y/X"""
        return chaospy.dist.operators.mul(self, X**-1)

    def __rfloordiv__(self, X):
        """Y.__rfloordiv__(X) <==> X/Y"""
        return chaospy.dist.operators.mul(X, self**-1)

    def __truediv__(self, X):
        """Y.__truediv__(X) <==> Y/X"""
        return chaospy.dist.operators.mul(self, X**-1)

    def __rtruediv__(self, X):
        """Y.__rtruediv__(X) <==> X/Y"""
        return chaospy.dist.operators.mul(X, self**-1)

    def __pow__(self, X):
        """Y.__pow__(X) <==> Y**X"""
        return chaospy.dist.operators.pow(self, X)

    def __rpow__(self, X):
        """Y.__rpow__(X) <==> X**Y"""
        return chaospy.dist.operators.pow(X, self)

    def __le__(self, X):
        """Y.__le__(X) <==> Y<=X"""
        return chaospy.dist.operators.trunk(self, X)

    def __lt__(self, X):
        """Y.__lt__(X) <==> Y<X"""
        return chaospy.dist.operators.trunk(self, X)

    def __ge__(self, X):
        """Y.__ge__(X) <==> Y>=X"""
        return chaospy.dist.operators.trunk(X, self)

    def __gt__(self, X):
        """Y.__gt__(X) <==> Y>X"""
        return chaospy.dist.operators.trunk(X, self)

    def addattr(self, **kws):
        """
        Add attribution to distribution

        Kwargs:
            pdf (callable) : Probability density function.
            cdf (callable) : Cumulative distribution function.
            ppf (callable) : Point percentile function.
            mom (callable) : Raw statistical moments.
            ttr (callable) : Three term recursion coefficient generator.
            val (callable) : If auxiliary variable, try to return the values
                    of it's underlying variables, else return self.
            str (callable, str) : Pretty print of module.
            dep (callable) : Dependency structure (if non-trivial).
        """
        for key,val in kws.items():
            if key=="str" and isinstance(val, str):
                val_ = val
                val = lambda *a,**k: val_
            setattr(self, "_"+key, types.MethodType(val, self))


    def dependent(self, *args):
        """
        Determine dependency structure in module

        Args:
            arg, [...] (optional) : If omited, internal dependency will be
                    determined.  If included, dependency between self and args
                    will be used.

        Returns:
            (bool) : True if distribution is dependent.
        """
        sets, graph = self.graph.run(None, "dep")

        if args:

            sets_ = set()
            for set_ in sets:
                sets_ = sets_.union(set_)
            sets = [sets_]

            for arg in args:
                sets_ = set()
                for set_ in arg.graph.run(None, "dep"):
                    sets_ = sets_.union(set_)
                sets.append(sets_)

        as_seperated = sum([len(set_) for set_ in sets])

        as_joined = set()
        for set_ in sets:
            as_joined = as_joined.union(set_)
        as_joined = len(as_joined)

        return as_seperated != as_joined


def construct(cdf, bnd, parent=None, pdf=None, ppf=None, mom=None, ttr=None,
              val=None, doc=None, str=None, dep=None, defaults=None,
              advance=False, length=1):
    """
    Random variable constructor.

    Args:
        cdf (callable) : Cumulative distribution function. Optional if parent
                is used.
        bnd (callable) : Boundary interval. Optional if parent is used.
        parent (Dist) : Distribution used as basis for new distribution. Any
                other argument that is omitted will instead take is function
                from parent.
        doc (str, optional) : Documentation for the distribution.
        str (str, callable, optional) : Pretty print of the variable.
        pdf (callable, optional) : Probability density function.
        ppf (callable, optional) : Point percentile function.
        mom (callable, optional) : Raw moment generator.
        ttr (callable, optional) : Three terms recursion coefficient generator
        val (callable, optional) : Value function for transferable
                distributions.
        dep (callable, optional) : Dependency structure.
        advance (bool) : If True, advance mode is used. See dist.graph for
                details.
        length (int) : If constructing an multivariate random variable, this
                sets the assumed length. Defaults to 1.
        init (callable, optional) : Custom constructor method.

    Returns:
        dist (Dist) : New custom distribution.
    """
    if not (parent is None):
        if hasattr(parent, "_cdf"):
            cdf = cdf or parent._cdf
        if hasattr(parent, "_bnd"):
            bnd = bnd or parent._bnd
        if hasattr(parent, "_pdf"):
            pdf = pdf or parent._pdf
        if hasattr(parent, "_ppf"):
            ppf = ppf or parent._ppf
        if hasattr(parent, "_mom"):
            mom = mom or parent._mom
        if hasattr(parent, "_ttr"):
            ttr = ttr or parent._ttr
        if hasattr(parent, "_str"):
            str = str or parent._str
        if hasattr(parent, "_dep"):
            dep = dep or parent._dep
        val = val or parent._val
        doc = doc or parent.__doc__

    def crash_func(*a, **kw):
        raise NotImplementedError
    if advance:
        ppf = ppf or crash_func
        pdf = pdf or crash_func
        mom = mom or crash_func
        ttr = ttr or crash_func

    def custom(**kws):

        if not (defaults is None):
            keys = defaults.keys()
            assert all([key in keys for key in kws.keys()])
            prm = defaults.copy()
        else:
            prm = {}
        prm.update(kws)
        _length = prm.pop("_length", length)
        _advance = prm.pop("_advance", advance)

        dist = Dist(_advance=_advance, _length=_length, **prm)

        dist.addattr(cdf=cdf)
        dist.addattr(bnd=bnd)

        if not (pdf is None):
            dist.addattr(pdf=pdf)
        if not (ppf is None):
            dist.addattr(ppf=ppf)
        if not (mom is None):
            dist.addattr(mom=mom)
        if not (ttr is None):
            dist.addattr(ttr=ttr)
        if not (val is None):
            dist.addattr(val=val)
        if not (str is None):
            dist.addattr(str=str)
        if not (dep is None):
            dist.addattr(dep=dep)

        return dist

    if not (doc is None):
        doc = """
Custom random variable
        """
    setattr(custom, "__doc__", doc)

    return custom



import numpy as np

import chaospy.dist


def samplegen(order, domain, rule="S", antithetic=None,
        verbose=False):
    """
    Sample generator.

    Interpretation of the domain argument:

    +------------+------------------------------------------------------------+
    | Value      | Interpretation                                             |
    +============+============================================================+
    | Dist       | Mapped to distribution domain using inverse Rosenblatt.    |
    +------------+------------------------------------------------------------+
    | int        | No mapping, but sets the number of dimension.              |
    +------------+------------------------------------------------------------+
    | array_like | Stretch samples such that they are in domain[0], domain[1] |
    +------------+------------------------------------------------------------+

    Intepretation of the rule argument:

    +------+---------------------+--------+
    | Key  | Name                | Nested |
    +======+=====================+========+
    | "C"  | Chebyshev nodes     | no     |
    +------+---------------------+--------+
    | "NC" | Nested Chebyshev    | yes    |
    +------+---------------------+--------+
    | "K"  | Korobov             | no     |
    +------+---------------------+--------+
    | "R"  | (Pseudo-)Random     | no     |
    +------+---------------------+--------+
    | "RG" | Regular grid        | no     |
    +------+---------------------+--------+
    | "NG" | Nested grid         | yes    |
    +------+---------------------+--------+
    | "L"  | Latin hypercube     | no     |
    +------+---------------------+--------+
    | "S"  | Sobol               | yes    |
    +------+---------------------+--------+
    | "H"  | Halton              | yes    |
    +------+---------------------+--------+
    | "M"  | Hammersley          | yes    |
    +------+---------------------+--------+

    Args:
        order (int) : Sample order.
        domain (Dist, int, array_like) : Defines the space where the samples
                are generated.
        rule (str) : rule for generating samples, where d is the number of
                dimensions.
        antithetic (array_like, optional) : List of bool. Represents the axes
                to mirror using antithetic variable.

    Examples:
        >>> print(cp.samplegen(3, cp.Normal(), "H"))
        [-2.33441422 -0.74196378  0.74196378  2.33441422]

        >>> cp.seed(1000)
        >>> print(cp.samplegen(3, cp.Normal(), "L"))
        [[ 0.6633974   0.46811863]
         [ 0.27875174  0.05308317]
         [ 0.98757072  0.51017741]
         [ 0.12054785  0.84929862]]
    """
    rule = rule.upper()

    if isinstance(domain, int):
        dim = domain
        trans = lambda x, verbose:x

    elif isinstance(domain, (tuple, list, np.ndarray)):
        domain = np.asfarray(domain)
        if len(domain.shape)<2:
            dim = 1
        else:
            dim = len(domain[0])
        lo,up = domain
        trans = lambda x, verbose: ((up-lo)*x.T + lo).T

    else:
        dist = domain
        dim = len(dist)
        trans = dist.inv

    if not (antithetic is None):

        antithetic = np.array(antithetic, dtype=bool).flatten()
        if antithetic.size==1 and dim>1:
            antithetic = np.repeat(antithetic, dim)

        N = np.sum(1*np.array(antithetic))
        order_,order = order,int(order*2**-N+1*(order%2!=0))
        trans_ = trans
        trans = lambda X, verbose: \
                trans_(antithetic_gen(X, antithetic)[:,:order_])

    if rule=="C":
        X = chaospy.dist.samplers.chebyshev(dim, order)

    elif rule=="NC":
        X = chaospy.dist.samplers.chebyshev_nested(dim, order)

    elif rule=="K":
        X = chaospy.dist.samplers.korobov(dim, order)

    elif rule=="R":
        X = np.random.random((dim,order))

    elif rule=="RG":
        X = chaospy.dist.samplers.regular_grid(dim, order)

    elif rule=="NG":
        X = chaospy.dist.samplers.regular_grid_nested(dim, order)

    elif rule=="L":
        X = chaospy.dist.samplers.latin_hypercube(dim, order)

    elif rule=="S":
        X = chaospy.dist.sobol_lib.sobol(dim, order)

    elif rule=="H":
        X = chaospy.dist.samplers.halton(dim, order)

    elif rule=="M":
        X = chaospy.dist.samplers.hammersley(dim, order)

    else:
        raise KeyError("rule not recognised")

    X = trans(X, verbose=verbose)

    return X



import inspect

import numpy
import chaospy.quad


def get_function(rule, domain, **parameters):
    """
    Create a quadrature function and set default parameter values.

    Args:
        rule (str) : Name of quadrature rule defined in `QUAD_FUNCTIONS`.
        domain (Dist, array_like) : Defines `lower` and `upper` that is passed
            quadrature rule. If `Dist`, `domain` is renamed to `dist` and also
            passed.
        **parameters (optional) : Redefining of the parameter defaults. Only
            add parameters that the quadrature rule expect.
    Returns:
        (callable) : Function that can be called only using argument `order`.
    """
    if isinstance(domain, chaospy.dist.Dist):
        lower, upper = domain.range()
        parameters["dist"] = domain
    else:
        lower, upper = numpy.array(domain)
    parameters["lower"] = lower
    parameters["upper"] = upper

    quad_function = chaospy.quad.collection.QUAD_FUNCTIONS[rule]
    parameters_spec = inspect.getargspec(quad_function)[0]
    parameters_spec = {key: None for key in parameters_spec}
    del parameters_spec["order"]

    for key in parameters_spec:
        if key in parameters:
            parameters_spec[key] = parameters[key]

    def _quad_function(order, *args, **kws):
        """Implementation of quadrature function."""
        params = parameters_spec.copy()
        params.update(kws)
        return quad_function(order, *args, **params)

    return _quad_function



"""
Collection of quadrature rules.

"Gaussian", "G"     Optimal Gaussian quadrature from Golub-Welsch
                    Slow for high order.
"Legendre", "E"     Gauss-Legendre quadrature
"Clenshaw", "C"     Clenshaw-Curtis quadrature. Exponential growth rule is
                    used when sparse is True to make the rule nested.
"Leja", J"          Leja quadrature. Linear growth rule is nested.
"Genz", "Z"         Hermite Genz-Keizter 16 rule. Nested. Valid to order 8.
"Patterson", "P"    Gauss-Patterson quadrature rule. Nested. Valid to order 8.
"""

from .interface import get_function

from .clenshaw_curtis import quad_clenshaw_curtis
from .gauss_patterson import quad_gauss_patterson
from .gauss_legendre import quad_gauss_legendre
from .genz_keister import quad_genz_keister
from .golub_welsch import quad_golub_welsch
from .leja import quad_leja

QUAD_FUNCTIONS = {
    "clenshaw_curtis": quad_clenshaw_curtis,
    "gauss_legendre": quad_gauss_legendre,
    "gauss_patterson": quad_gauss_patterson,
    "genz_keister": quad_genz_keister,
    "golub_welsch": quad_golub_welsch,
    "leja": quad_leja,
}

QUAD_SHORT_NAMES = {
    "c": "clenshaw_curtis",
    "e": "gauss_legendre",
    "p": "gauss_patterson",
    "z": "genz_keister",
    "g": "golub_welsch",
    "j": "leja",
}



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name='chaospy',
    version="2.2.0",
    url='https://github.com/jonathf/chaospy',
    author="Jonathan Feinberg",
    author_email="jonathf@gmail.com",
    license='BSD',
    platforms='any',
    packages=find_packages("src"),
    package_dir={"": "src"},
    install_requires=[
        "numpy", "scipy", "networkx"
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Mathematics',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
    ],
)



"""
Multiplication.
"""
import numpy

from chaospy.dist import Dist
import chaospy.dist.operators


class Mul(Dist):
    """Multiplication."""

    def __init__(self, left, right):
        """
        Constructor.

        Args:
            left (Dist, array_like) : Left hand side.
            right (Dist, array_like) : Right hand side.
        """
        left_ = not isinstance(left, Dist) or 1 and len(left)
        right_ = not isinstance(right, Dist) or 1 and len(right)
        length = max(left_, right_)
        Dist.__init__(
            self, left=left, right=right, _length=length, _advance=True)


    def _bnd(self, xloc, graph):
        """Distribution bounds."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]

        num = (num + 1.*(num == 0))
        xloc = (xloc.T/num.T).T
        lower_, upper_ = graph(xloc, dist)
        lower = (num.T*(lower_.T*(num.T > 0) + upper_.T*(num.T < 0))).T
        upper = (num.T*(upper_.T*(num.T > 0) + lower_.T*(num.T < 0))).T

        return lower, upper

    def _cdf(self, xloc, graph):
        """Cumulative distribution function."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
        xloc = (xloc.T/(num.T + 1.*(num.T == 0))).T
        out = graph(xloc, dist)
        out = (out.T*(num.T > 0) + (1.-out.T)*(num.T < 0) + \
                1.*(xloc.T > num.T)*(num.T == 0)).T
        return out

    def _ppf(self, uloc, graph):
        """Point percentile function."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
        uloc = (uloc*(num > 0) + (1.-uloc)*(num <= 0))
        return graph(uloc, dist)*num

    def _pdf(self, xloc, graph):
        """Probability density function."""
        if "left" in graph.keys and "right" in graph.dists:
            num, dist = graph.keys["left"], graph.dists["right"]
        else:
            num, dist = graph.keys["right"], graph.dists["left"]
        num = numpy.where(num, num, numpy.inf)
        return numpy.abs(graph(xloc*1./num, dist)/num)

    def _mom(self, key, graph):
        """Statistical moments."""
        if len(graph.dists) == 2 and\
                graph.dists["left"].dependent(graph.dists["right"]):
            raise NotImplementedError("dependency")

        left = []
        if "left" in graph.dists:
            left.append(graph(key, graph.dists["left"]))
        else:
            left.append((graph.keys["left"].T**key.T).T)
        if "right" in graph.dists:
            left.append(graph(key, graph.dists["right"]))
        else:
            left.append((graph.keys["right"].T**key.T).T)

        return numpy.prod(left[0]*left[1], 0)

    def _ttr(self, order, graph):
        """Three terms recursion coefficients."""
        if "left" in graph.keys and "right" in graph.dists:
            num = graph.keys["left"]
            dist = graph.dists["right"]
        else:
            dist = graph.dists["left"]
            num = graph.keys["right"]

        coeff0, coeff1 = graph(order, dist)
        return coeff0*num, coeff1*num*num

    def _val(self, graph):
        """Value extraction."""
        if len(graph.keys) == 2:
            return graph.keys["left"]*graph.keys["right"]
        return self

    def _str(self, left, right):
        """String representation."""
        return str(left) + "*" + str(right)


class Mvmul(Dist):
    """Multiplication for multivariate variables."""

    def __init__(self, dist, C):
        """
        Constructor.

        Args:
            dist (Dist, array_like) : Probability.
            C (numpy.ndarray) : matrix to multiply with.
        """
        Dist.__init__(self, dist=dist, C=C,
                Ci=numpy.linalg.inv(C),
                _length=len(dist), _advance=True)

    def _cdf(self, xloc, graph):
        """Cumulative distribution function."""
        return graph(numpy.dot(graph.keys["Ci"], xloc), graph.dists["dist"])

    def _ppf(self, q, graph):
        """Point percentile function."""
        return numpy.dot(graph.keys["C"], graph(q, graph.dists["dist"]))

    def _bnd(self, xloc, graph):
        """Distribution bounds."""
        bnd = graph(xloc, graph.dists["dist"])
        C = graph.keys["C"]
        lower = (numpy.dot(C, bnd[0]).T).T
        upper = (numpy.dot(C, bnd[1]).T).T

        wrong = lower>upper
        out = numpy.where(wrong, upper, lower), numpy.where(wrong, lower, upper)
        return out

    def _val(self, graph):
        """Value extraction."""
        if "dist" in graph.keys:
            return numpy.dot(graph.keys["dist"].T, graph.keys["C"].T).T
        return self

    def _str(self, C, Ci, dist):
        """String representation."""
        return str(dist) + "*" + str(C)

    def _dep(self, graph):
        """Dependency evaluation."""
        dist = graph.dists["dist"]
        S = graph(dist)

        out = [set([]) for _ in range(len(self))]
        C = graph.keys["C"]

        for i in range(len(self)):
            for j in range(len(self)):
                if C[i,j]:
                    out[i].update(S[j])

        return out


def mul(left, right):
    """
    Distribution multiplication.

    Args:
        left (Dist, array_like) : left hand side.
        right (Dist, array_like) : right hand side.
    """
    if left is right:
        return pow(left, 2)

    if isinstance(left, Dist):

        if not isinstance(right, Dist):
            right = numpy.array(right)
            if right.size == 1:
                if right == 1:
                    return left
                if right == 0:
                    return 0.

    elif isinstance(right, Dist):

        left = numpy.array(left)
        if left.size == 1:
            if left == 1:
                return right
            if left == 0:
                return 0.

    else:
        return left*right

    a = not isinstance(left, Dist) or 1 and len(left)
    b = not isinstance(right, Dist) or 1 and len(right)
    length = max(a, b)
    if length == 1:
        return Mul(left, right)
    return Mvmul(dist=left, C=right)



r"""
In practice the following four components are needed to perform psuedo-spectral
projection. (For the "real" spectral projection method, see: :ref:`galerkin`):

-  A distribution for the unknown function parameters (as described in
   section :ref:`distributions`). For example::

      >>> distribution = chaospy.Iid(chaospy.Normal(0, 1), 2)

-  Create integration absissas and weights (as described in :ref:`quadrature`)::

      >>> absissas, weights = chaospy.generate_quadrature(
      ...     2, distribution, rule="G")
      >>> print(numpy.around(absissas, 15))
      [[-1.73205081 -1.73205081 -1.73205081  0.          0.          0.
         1.73205081  1.73205081  1.73205081]
       [-1.73205081  0.          1.73205081 -1.73205081  0.          1.73205081
        -1.73205081  0.          1.73205081]]
      >>> print(weights)
      [ 0.02777778  0.11111111  0.02777778  0.11111111  0.44444444  0.11111111
        0.02777778  0.11111111  0.02777778]

- An orthogonal polynomial expansion (as described in section
  :ref:`orthogonality`) where the weight function is the distribution in the
  first step::

      >>> orthogonal_expansion = chaospy.orth_ttr(2, distribution)
      >>> print(orthogonal_expansion)
      [1.0, q1, q0, q1^2-1.0, q0q1, q0^2-1.0]

- A function evaluated using the nodes generated in the second step.
  For example::

      >>> def model_solver(q):
      ...     return [q[0]*q[1], q[0]*numpy.e**-q[1]+1]
      >>> solves = [model_solver(absissa) for absissa in absissas.T]
      >>> print(numpy.around(solves[:4], 8))
      [[ 3.         -8.7899559 ]
       [-0.         -0.73205081]
       [-3.          0.69356348]
       [-0.          1.        ]]

- To bring it together, expansion, absissas, weights and solves are used as
  arguments to create approximation::

      >>> approx = chaospy.fit_quadrature(
      ...     orthogonal_expansion, absissas, weights, solves)
      >>> print(chaospy.around(approx, 8))
      [q0q1, -1.58058656q0q1+1.63819248q0+1.0]

Note that in this case the function output is
bivariate. The software is designed to create an approximation of any
discretized model as long as it is compatible with ``numpy`` shapes.

As mentioned in section :ref:`orthogonality`, moment based construction of
polynomials can be unstable. This might also be the case for the
denominator :math:`\mathbb E{\Phi_n^2}`. So when using three terms
recursion, it is common to use the recurrence coefficients to estimated
the denominator.

One cavat with using psuedo-spectral projection is that the calculations of the
norms of the polynomials becomes unstable. To mittigate, recurrence
coefficients can be used to calculate them instead with more stability.
To include these stable norms in the calculations, the following change in code
can be added::

   >>> orthogonal_expansion, norms = chaospy.orth_ttr(2, distribution, retall=True)
   >>> approx2 = chaospy.fit_quadrature(
   ...     orthogonal_expansion, absissas, weights, solves, norms=norms)
   >>> print(chaospy.around(approx2, 8))
   [q0q1, -1.58058656q0q1+1.63819248q0+1.0]

Note that at low polynomial order, the error is very small. For example the
largest coefficient between the two approximation::

   >>> print(numpy.max(abs(approx-approx2).coeffs(), -1) < 1e-12)
   [ True  True]

The ``coeffs`` function returns all the polynomial coefficients.
"""

import numpy

import chaospy


def pcm(func, porder, dist, rule="G", sorder=None, proxy_dist=None,
        orth=None, orth_acc=100, quad_acc=100, sparse=False, composit=1,
        antithetic=None, lr="LS", **kws):
    """
Probabilistic Collocation Method.

Args:
    func (callable) : The model to be approximated.  Must accept arguments on
            the form `func(z, *args, **kws)` where `z` is an 1-dimensional
            array with `len(z)==len(dist)`.
    porder (int) : The order of the polynomial approximation.
    dist (Dist) : Distributions for models parameter.
    rule (str) : The rule for estimating the Fourier coefficients.  For
            spectral projection/quadrature rules, see generate_quadrature.
            For point collocation/nummerical sampling, see samplegen.

Kwargs:
    proxy_dist (Dist) : If included, the expansion will be created in
            proxy_dist and values will be mapped to dist using a double
            Rosenblatt transformation.
    sorder (float) : The order of the sample scheme used.  If omited, default
            values will be used.
    orth (callable, Poly) : Orthogonal polynomial generation.
            If callable, the return of orth(order, dist) will be used. If Poly
            is provided, it will be used directly. If omited, orthogonal
            polynomial will be `orth_ttr` or `orth_chol` depending on if
            `dist` is stochastically independent of not.
    orth_acc (int) : Accuracy used in the estimation of polynomial expansion.
    sparse (bool) : If True, Smolyak sparsegrid will be used instead of full
            tensorgrid.
    composit (int) : Use composit rule. Note that the number of evaluations
            may grow quickly.
    antithetic (bool, array_like) : Use of antithetic variable
    lr (str) : Linear regresion method.
    lr_kws (dict) : Extra keyword arguments passed to fit_regression.

Returns:
    q : Poly
        Polynomial approximation of a given a model.

Examples:
    >>> func = lambda z: -z[1]**2 + 0.1*z[0]
    >>> dist = chaospy.J(chaospy.Uniform(), chaospy.Uniform())
    >>> q = chaospy.pcm(func, 2, dist)
    >>> print(chaospy.around(q, 10))
    -q1^2+0.1q0
    """
    # Proxy variable
    if proxy_dist is None:
        trans = lambda x:x
    else:
        dist, dist_ = proxy_dist, dist
        trans = lambda x: dist_.inv(dist.fwd(x), **kws)

    # The polynomial expansion
    if orth is None:
        if dist.dependent():
            orth = chaospy.orthogonal.orth_chol
        else:
            orth = chaospy.orthogonal.orth_ttr
    if not isinstance(orth, chaospy.poly.Poly):
        orth = orth(porder, dist, acc=orth_acc, **kws)

    # Applying scheme
    rule = rule.upper()
    if rule in "GEC":
        if sorder is None:
            sorder = porder+1
        z, w = chaospy.quad.generate_quadrature(
            sorder, dist, acc=quad_acc, sparse=sparse, rule=rule,
            composit=composit, **kws)

        x = trans(z)
        y = numpy.array([func(_) for _ in x.T])
        Q = fit_quadrature(orth, x, w, y, **kws)

    else:
        if sorder is None:
            sorder = 2*len(orth)
        z = dist.sample(sorder, rule=rule, antithetic=antithetic)

        x = trans(z)
        y = numpy.array(map(func, x.T))
        Q = fit_regression(orth, x, y, rule=lr, **kws)

    return Q


def pcm_cc(func, order, dist_out, dist_in=None, acc=None,
        orth=None, retall=False, sparse=False):
    """
Probabilistic Collocation Method using Clenshaw-Curtis quadrature

Parameters
----------
Required arguments

func : callable
    The model to be approximated.
    Must accept arguments on the form `func(z, *args, **kws)`
    where `z` is an 1-dimensional array with `len(z)==len(dist)`.
order : int
    The order of the polynomial approximation
dist_out : Dist
    Distributions for models parameter

Optional arguments

dist_in : Dist
    If included, space will be mapped using a Rosenblatt
    transformation from dist_out to dist_in before creating an
    expansin in terms of dist_in
acc : float
    The order of the sample scheme used
    If omitted order+1 will be used
orth : int, str, callable, Poly
    Orthogonal polynomial generation.

    int, str :
        orth will be passed to orth_select
        for selection of orthogonalization.
        See orth_select doc for more details.

    callable :
        the return of orth(M, dist) will be used.

    Poly :
        it will be used directly.
        All polynomials must be orthogonal for method to work
        properly.
retall : bool
    If True, return extra values.
sparse : bool
    If True, Smolyak sparsegrid will be used instead of full
    tensorgrid

Returns
-------
q[, x, w, y]

q : Poly
    Polynomial estimate of a given a model.
x : numpy.ndarray
    Nodes used in quadrature with `x.shape=(dim, K)` where K is the
    number of samples.
w : numpy.ndarray
    Weights used in quadrature with `w.shape=(K,)`.
y : numpy.ndarray
    Evauluations of func with `len(y)=K`.

#  Examples
#  --------
#
#  Define function and distribution:
#  >>> func = lambda z: -z[1]**2 + 0.1*z[0]
#  >>> dist = chaospy.J(chaospy.Uniform(), chaospy.Uniform())
#
#  Perform pcm:
#  >>> q, x, w, y = chaospy.pcm_cc(func, 2, dist, acc=2, retall=1)
#  >>> print(chaospy.around(q, 10))
#  -q1^2+0.1q0
#  >>> print(len(w))
#  9
#
#  With Smolyak sparsegrid
#  >>> q, x, w, y = chaospy.pcm_cc(func, 2, dist, acc=2, retall=1, sparse=1)
#  >>> print(chaospy.around(q, 10))
#  -q1^2+0.1q0
#  >>> print(len(w))
#  13
    """
    if acc is None:
        acc = order+1

    if dist_in is None:
        z, w = chaospy.quadrature.generate_quadrature(
            acc, dist_out, 100, sparse=sparse, rule="C")
        x = z
        dist = dist_out
    else:
        z, w = chaospy.quadrature.generate_quadrature(
            acc, dist_in, 100, sparse=sparse, rule="C")
        x = dist_out.ppf(dist_in.cdf(z))
        dist = dist_in

    if orth is None:
        if dist.dependent:
            orth = "chol"
        else:
            orth = "ttr"
    if isinstance(orth, (str, int)):
        orth = orth_select(orth)
    if not isinstance(orth, chaospy.poly.Poly):
        orth = orth(order, dist)

    y = numpy.array(map(func, x.T))
    Q = fit_quadrature(orth, x, w, y)

    if retall:
        return Q, x, w, y
    return Q

def fit_quadrature(orth, nodes, weights, solves, retall=False,
        norms=None, **kws):
    """
Using spectral projection to create a polynomial approximation over
distribution space.

Parameters
----------
orth : Poly
    Orthogonal polynomial expansion. Must be orthogonal for the
    approximation to be accurate.
nodes : array_like
    Where to evaluate the polynomial expansion and model to
    approximate.
    nodes.shape==(D,K) where D is the number of dimensions and K is
    the number of nodes.
weights : array_like
    Weights when doing numerical integration.
    weights.shape==(K,)
solves : array_like, callable
    The model to approximate.
    If array_like is provided, it must have len(solves)==K.
    If callable, it must take a single argument X with len(X)==D,
    and return a consistent numpy compatible shape.
norms : array_like
    In the of TTR using coefficients to estimate the polynomial
    norm is more stable than manual calculation.
    Calculated using quadrature if no provided.
    norms.shape==(len(orth),)
    """

    orth = chaospy.poly.Poly(orth)
    nodes = numpy.asfarray(nodes)
    weights = numpy.asfarray(weights)

    if hasattr(solves, "__call__"):
        solves = [solves(q) for q in nodes.T]
    solves = numpy.asfarray(solves)

    shape = solves.shape
    solves = solves.reshape(weights.size, solves.size/weights.size)

    ovals = orth(*nodes)
    vals1 = [(val*solves.T*weights).T for val in ovals]

    if norms is None:
        vals2 = [(val**2*weights).T for val in ovals]
        norms = numpy.sum(vals2, 1)
    else:
        norms = numpy.array(norms).flatten()
        assert len(norms)==len(orth)

    coefs = (numpy.sum(vals1, 1).T/norms).T
    coefs = coefs.reshape(len(coefs), *shape[1:])
    Q = chaospy.poly.transpose(chaospy.poly.sum(orth*coefs.T, -1))

    if retall:
        return Q, coefs
    return Q



def pcm_gq(func, order, dist_out, dist_in=None, acc=None,
        orth=None, retall=False, sparse=False):
    """
Probabilistic Collocation Method using optimal Gaussian quadrature

Parameters
----------
Required arguments

func : callable
    The model to be approximated.
    Must accept arguments on the form `func(z, *args, **kws)`
    where `z` is an 1-dimensional array with `len(z)==len(dist)`.
order : int
    The order of the polynomial approximation
dist_out : Dist
    Distributions for models parameter

Optional arguments

dist_in : Dist
    If included, space will be mapped using a Rosenblatt
    transformation from dist_out to dist_in before creating an
    expansin in terms of dist_in
acc : float
    The order of the sample scheme used
    If omitted order+1 will be used
orth : int, str, callable, Poly
    Orthogonal polynomial generation.

    int, str :
        orth will be passed to orth_select
        for selection of orthogonalization.
        See orth_select doc for more details.

    callable :
        the return of orth(M, dist) will be used.

    Poly :
        it will be used directly.
        All polynomials must be orthogonal for method to work
        properly.
args : itterable
    Extra positional arguments passed to `func`.
kws : dict
    Extra keyword arguments passed to `func`.
retall : bool
    If True, return also number of evaluations
sparse : bool
    If True, Smolyak sparsegrid will be used instead of full
    tensorgrid

Returns
-------
Q[, X]

Q : Poly
    Polynomial estimate of a given a model.
X : numpy.ndarray
    Values used in evaluation

#  Examples
#  --------
#  Define function:
#  >>> func = lambda z: z[1]*z[0]
#
#  Define distribution:
#  >>> dist = chaospy.J(chaospy.Normal(), chaospy.Normal())
#
#  Perform pcm:
#  >>> p, x, w, y = chaospy.pcm_gq(func, 2, dist, acc=3, retall=True)
#  >>> print(chaospy.around(p, 10))
#  q0q1
#  >>> print(len(w))
#  16

    """
    if acc is None:
        acc = order+1

    if dist_in is None:
        z, w = chaospy.quadrature.generate_quadrature(
            acc, dist_out, 100, sparse=sparse, rule="G")
        x = z
        dist = dist_out
    else:
        z, w = chaospy.quadrature.generate_quadrature(
            acc, dist_in, 100, sparse=sparse, rule="G")
        x = dist_out.ppf(dist_in.cdf(z))
        dist = dist_in

    y = numpy.array(map(func, x.T))
    shape = y.shape
    y = y.reshape(w.size, y.size/w.size)

    if orth is None:
        if dist.dependent:
            orth = "chol"
        else:
            orth = "ttr"
    if isinstance(orth, (str, int)):
        orth = orth_select(orth)
    if not isinstance(orth, chaospy.poly.Poly):
        orth = orth(order, dist)

    ovals = orth(*z)
    vals1 = [(val*y.T*w).T for val in ovals]
    vals2 = [(val**2*w).T for val in ovals]
    coef = (numpy.sum(vals1, 1).T/numpy.sum(vals2, 1)).T

    coef = coef.reshape(len(coef), *shape[1:])
    Q = chaospy.poly.transpose(chaospy.poly.sum(orth*coef.T, -1))

    if retall:
        return Q, x, w, y
    return Q


def pcm_lr(func, order, dist_out, sample=None,
        dist_in=None, rule="H",
        orth=3, regression="LS", retall=False):
    """
Probabilistic Collocation Method using Linear Least Squares fit

Parameters
----------
Required arguemnts

func : callable
    The model to be approximated.  Must accept arguments on the
    form `z` is an 1-dimensional array with `len(z)==len(dist)`.
order : int
    The order of chaos expansion.
dist_out : Dist
    Distributions for models parameter.

Optional arguments

sample : int
    The order of the sample scheme to be used.
    If omited it defaults to 2*len(orth).
dist_in : Dist
    If included, space will be mapped using a Rosenblatt
    transformation from dist_out to dist_in before creating an
    expansin in terms of dist_in
rule:
    rule for generating samples, where d is the number of
    dimensions.

    Key     Name                Nested
    ----    ----------------    ------
    "K"     Korobov             no
    "R"     (Pseudo-)Random     no
    "L"     Latin hypercube     no
    "S"     Sobol               yes
    "H"     Halton              yes
    "M"     Hammersley          yes

orth : int, str, callable, Poly
    Orthogonal polynomial generation.

    int, str :
        orth will be passed to orth_select
        for selection of orthogonalization.
        See orth_select doc for more details.

    callable :
        the return of orth(M, dist) will be used.

    Poly :
        it will be used directly.
        It must be of length N+1=comb(M+D, M)
regression : str
    Linear regression method used.
    See fit_regression for more details.
retall : bool
    If True, return extra values.

#  Examples
#  --------
#
#  Define function:
#  >>> func = lambda z: -z[1]**2 + 0.1*z[0]
#
#  Define distribution:
#  >>> dist = chaospy.J(chaospy.Normal(), chaospy.Normal())
#
#  Perform pcm:
#  >>> q, x, y = chaospy.pcm_lr(func, 2, dist, retall=True)
#  >>> print(chaospy.around(q, 10))
#  -q1^2+0.1q0
#  >>> print(len(x.T))
#  12
    """

    if dist_in is None:
        dist = dist_out
    else:
        dist = dist_in

    # orthogonalization
    if orth is None:
        if dist.dependent():
            orth = "chol"
        else:
            orth = "ttr"
    if isinstance(orth, (str, int)):
        orth = orth_select(orth)
    if not isinstance(orth, chaospy.poly.Poly):
        orth = orth(order, dist)

    # sampling
    if sample is None:
        sample = 2*len(orth)

    x = chaospy.dist.samplegen(sample, dist, rule)


    # Rosenblatt
    if not (dist_in is None):
        x = dist_out.ppf(dist_in.cdf(x))

    # evals
    y = numpy.array(map(func, x.T))
    shape = y.shape[1:]
    y = y.reshape(len(y), y.size/len(y))
    if sample==0:
        y_ = y[:]
        R = orth * y
    else:
        R, y_ = fit_regression(orth, x, y, regression, retall=1)

    R = chaospy.poly.reshape(R, shape)

    if retall:
        return R, x, y
    return R




def fit_lagrange(X, Y):
    """Simple lagrange method"""
    X = numpy.array(X)
    Y = numpy.array(Y)
    assert X.shape[0] == Y.shape[0]

    if len(X.shape) == 1:
        X = X.reshape(1, X.size)

    N, dim = X.shape

    basis = []
    n = 1
    while len(basis) < N:
        basis = chaospy.poly.basis(0, n, dim)
        n += 1

    basis = basis[:N]

    return fit_regression(basis, X, Y)



def cross_validate(P, X, Y, folds=None, rule="LS", **kws):
    """
Parameters
----------
P : Poly
    Polynomial expansion
X : array_like
    Input data with X.shape=(dim,K)
Y : array_like
    Output data with len(Y)=K
folds : int,optional
    Number of folds in validation. If omitted selected to be K.
    """
    X = numpy.array(X)
    dim,K = X.shape
    Y = numpy.array(Y)
    assert len(Y)==K

    out = Y.copy()

    if folds is None:
        folds = K
    R = numpy.random.randint(0, folds, K)

    for fold in range(folds):

        infold = R==fold
        x = X[:, True-infold]
        y = Y[True-infold]
        poly = fit_regression(P, x, y, rule=rule, **kws)
        out[infold] = out[infold]-poly(*X[:,infold])

    return out
